from abc import abstractmethod
from typing import Callable, Dict, List, Text, Tuple, Union

from aesara.graph.basic import Apply, Constant
from aesara.graph.utils import MethodNotDefined


class CLinkerObject:
    """Standard methods for an `Op` or `Type` used with the `CLinker`."""

    def c_headers(self, **kwargs) -> List[Text]:
        """Return a list of header files required by code returned by this class.

        These strings will be prefixed with ``#include`` and inserted at the
        beginning of the C source code.

        Strings in this list that start neither with ``<`` nor ``"`` will be
        enclosed in double-quotes.

        Examples
        --------

        .. code-block:: python

            def c_headers(self, **kwargs):
                return ['<iostream>', '<math.h>', '/full/path/to/header.h']


        """
        return []

    def c_header_dirs(self, **kwargs) -> List[Text]:
        """Return a list of header search paths required by code returned by this class.

        Provides search paths for headers, in addition to those in any relevant
        environment variables.

        .. note::

            For Unix compilers, these are the things that get ``-I`` prefixed
            in the compiler command line arguments.


        Examples
        --------

        .. code-block:: python

            def c_header_dirs(self, **kwargs):
                return ['/usr/local/include', '/opt/weirdpath/src/include']

        """
        return []

    def c_libraries(self, **kwargs) -> List[Text]:
        """Return a list of libraries required by code returned by this class.

        The compiler will search the directories specified by the environment
        variable ``LD_LIBRARY_PATH`` in addition to any returned by
        :meth:`CLinkerOp.c_lib_dirs`.

        .. note::

            For Unix compilers, these are the things that get ``-l`` prefixed
            in the compiler command line arguments.


        Examples
        --------

        .. code-block:: python

            def c_libraries(self, **kwargs):
                return ['gsl', 'gslcblas', 'm', 'fftw3', 'g2c'].

        """
        return []

    def c_lib_dirs(self, **kwargs) -> List[Text]:
        """Return a list of library search paths required by code returned by this class.

        Provides search paths for libraries, in addition to those in any
        relevant environment variables (e.g. ``LD_LIBRARY_PATH``).

        .. note::

            For Unix compilers, these are the things that get ``-L`` prefixed
            in the compiler command line arguments.


        Examples
        --------

        .. code-block:: python

            def c_lib_dirs(self, **kwargs):
                return ['/usr/local/lib', '/opt/weirdpath/build/libs'].

        """
        return []

    def c_support_code(self, **kwargs) -> Text:
        """Return utility code for use by a `Variable` or `Op`.

        This is included at global scope prior to the rest of the code for this class.

        Question: How many times will this support code be emitted for a graph
        with many instances of the same type?

        Returns
        -------
        str

        """
        return ""

    def c_compile_args(self, **kwargs) -> List[Text]:
        """Return a list of recommended compile arguments for code returned by other methods in this class.

        Compiler arguments related to headers, libraries and search paths
        should be provided via the functions `c_headers`, `c_libraries`,
        `c_header_dirs`, and `c_lib_dirs`.

        Examples
        --------

        .. code-block:: python

            def c_compile_args(self, **kwargs):
                return ['-ffast-math']

        """
        return []

    def c_no_compile_args(self, **kwargs) -> List[Text]:
        """Return a list of incompatible ``gcc`` compiler arguments.

        We will remove those arguments from the command line of ``gcc``. So if
        another `Op` adds a compile arg in the graph that is incompatible
        with this `Op`, the incompatible arg will not be used.

        This is used, for instance, to remove ``-ffast-math``.

        """
        return []

    def c_init_code(self, **kwargs) -> List[Text]:
        """Return a list of code snippets to be inserted in module initialization."""
        return []

    def c_code_cache_version(self) -> Union[Tuple[int], Tuple]:
        """Return a tuple of integers indicating the version of this `Op`.

        An empty tuple indicates an "unversioned" `Op` that will not be cached
        between processes.

        The cache mechanism may erase cached modules that have been superseded
        by newer versions. See `ModuleCache` for details.

        See Also
        --------
        c_code_cache_version_apply()

        """
        return ()


class CLinkerOp(CLinkerObject):
    """Interface definition for `Op` subclasses compiled by `CLinker`."""

    @abstractmethod
    def c_code(
        self,
        node: Apply,
        name: Text,
        inputs: List[Text],
        outputs: List[Text],
        sub: Dict[Text, Text],
    ) -> Text:
        """Return the C implementation of an ``Op``.

        Returns C code that does the computation associated to this ``Op``,
        given names for the inputs and outputs.

        Parameters
        ----------
        node : Apply instance
            The node for which we are compiling the current C code.
            The same ``Op`` may be used in more than one node.
        name : str
            A name that is automatically assigned and guaranteed to be
            unique.
        inputs : list of strings
            There is a string for each input of the function, and the
            string is the name of a C variable pointing to that input.
            The type of the variable depends on the declared type of
            the input.  There is a corresponding python variable that
            can be accessed by prepending ``"py_"`` to the name in the
            list.
        outputs : list of strings
            Each string is the name of a C variable where the `Op` should
            store its output.  The type depends on the declared type of
            the output.  There is a corresponding Python variable that
            can be accessed by prepending ``"py_"`` to the name in the
            list.  In some cases the outputs will be preallocated and
            the value of the variable may be pre-filled.  The value for
            an unallocated output is type-dependent.
        sub : dict of strings
            Extra symbols defined in `CLinker` sub symbols (such as ``'fail'``).

        """
        raise NotImplementedError()

    def c_code_cache_version_apply(self, node: Apply) -> Tuple[int]:
        """Return a tuple of integers indicating the version of this `Op`.

        An empty tuple indicates an "unversioned" `Op` that will not be
        cached between processes.

        The cache mechanism may erase cached modules that have been
        superseded by newer versions.  See `ModuleCache` for details.

        See Also
        --------
        c_code_cache_version

        Notes
        -----
            This function overrides `c_code_cache_version` unless it explicitly
            calls `c_code_cache_version`. The default implementation simply
            calls `c_code_cache_version` and ignores the `node` argument.

        """
        return self.c_code_cache_version()

    def c_code_cleanup(
        self,
        node: Apply,
        name: Text,
        inputs: List[Text],
        outputs: List[Text],
        sub: Dict[Text, Text],
    ) -> Text:
        """Return C code to run after :meth:`CLinkerOp.c_code`, whether it failed or not.

        This is a convenient place to clean up things allocated by :meth:`CLinkerOp.c_code`.

        Parameters
        ----------
        node : Apply
            WRITEME
        name : str
            A name that is automatically assigned and guaranteed to be
            unique.
        inputs : list of str
            There is a string for each input of the function, and the
            string is the name of a C variable pointing to that input.
            The type of the variable depends on the declared type of
            the input. There is a corresponding Python variable that
            can be accessed by prepending ``"py_"`` to the name in the
            list.
        outputs : list of str
            Each string is the name of a C variable corresponding to
            one of the outputs of the `Op`. The type depends on the
            declared type of the output. There is a corresponding
            Python variable that can be accessed by prepending ``"py_"`` to
            the name in the list.
        sub : dict of str
            Extra symbols defined in `CLinker` sub symbols (such as ``'fail'``).

        """
        return ""

    def c_support_code_apply(self, node: Apply, name: Text) -> Text:
        """Return `Apply`-specialized utility code for use by an `Op` that will be inserted at global scope.

        Parameters
        ----------
        node : Apply
            The node in the graph being compiled.
        name : str
            A string or number that serves to uniquely identify this node.
            Symbol names defined by this support code should include the name,
            so that they can be called from the :meth:`CLinkerOp.c_code`, and so that
            they do not cause name collisions.

        Notes
        -----
        This function is called in addition to :meth:`CLinkerObject.c_support_code`
        and will supplement whatever is returned from there.

        """
        return ""

    def c_init_code_apply(self, node: Apply, name: Text) -> Text:
        """Return a code string specific to the `Apply` to be inserted in the module initialization code.

        Parameters
        ----------
        node
            An `Apply` instance in the graph being compiled
        name : str
            A string or number that serves to uniquely identify this node.
            Symbol names defined by this support code should include the name,
            so that they can be called from :meth:`CLinkerOp.c_code`, and so
            that they do not cause name collisions.

        Notes
        -----
        This function is called in addition to
        :meth:`CLinkerObject.c_init_code` and will supplement whatever is
        returned from there.

        """
        return ""

    def c_init_code_struct(self, node: Apply, name, sub) -> Text:
        """Return an `Apply`-specific code string to be inserted in the struct initialization code.

        Parameters
        ----------
        node : Apply
            The node in the graph being compiled.
        name : str
            A unique name to distinguish variables from those of other nodes.
        sub : dict of str
            A dictionary of values to substitute in the code.
            Most notably it contains a ``'fail'`` entry that you should place
            in your code after setting a Python exception to indicate an error.

        """
        return ""

    def c_support_code_struct(self, node: Apply, name: Text) -> Text:
        """Return `Apply`-specific utility code for use by an `Op` that will be inserted at struct scope.

        Parameters
        ----------
        node : Apply
            The node in the graph being compiled
        name : str
            A unique name to distinguish you variables from those of other
            nodes.

        """
        return ""

    def c_cleanup_code_struct(self, node: Apply, name: Text) -> Text:
        """Return an `Apply`-specific code string to be inserted in the struct cleanup code.

        Parameters
        ----------
        node : Apply
            The node in the graph being compiled
        name : str
            A unique name to distinguish variables from those of other nodes.

        """
        return ""


class CLinkerType(CLinkerObject):
    r"""Interface specification for `Type`\s that can be arguments to a `CLinkerOp`.

    A `CLinkerType` instance is mainly responsible  for providing the C code that
    interfaces python objects with a C `CLinkerOp` implementation.

    """

    @abstractmethod
    def c_declare(
        self, name: Text, sub: Dict[Text, Text], check_input: bool = True
    ) -> Text:
        """Return C code to declare variables that will be instantiated by :meth:`CLinkerType.c_extract`.

        Parameters
        ----------
        name
            The name of the ``PyObject *`` pointer that will the value for this
            `Type`.
        sub
            A dictionary of special codes.  Most importantly
            ``sub['fail']``. See `CLinker` for more info on ``sub`` and
            ``fail``.

        Notes
        -----
        It is important to include the `name` inside of variables which
        are declared here, so that name collisions do not occur in the
        source file that is generated.

        The variable called `name` is not necessarily defined yet
        where this code is inserted. This code might be inserted to
        create class variables for example, whereas the variable `name`
        might only exist inside certain functions in that class.

        TODO: Why should variable declaration fail?  Is it even allowed to?

        Examples
        --------

        .. code-block: python

            def c_declare(self, name, sub, check_input=True):
                return "PyObject ** addr_of_%(name)s;"

        """

    @abstractmethod
    def c_init(self, name: Text, sub: Dict[Text, Text]) -> Text:
        """Return C code to initialize the variables that were declared by :meth:`CLinkerType.c_declare`.

        Notes
        -----
        The variable called `name` is not necessarily defined yet
        where this code is inserted. This code might be inserted in a
        class constructor for example, whereas the variable `name`
        might only exist inside certain functions in that class.

        TODO: Why should variable initialization fail?  Is it even allowed to?

        Examples
        --------

        .. code-block: python

            def c_init(self, name, sub):
                return "addr_of_%(name)s = NULL;"

        """

    @abstractmethod
    def c_extract(
        self, name: Text, sub: Dict[Text, Text], check_input: bool = True, **kwargs
    ) -> Text:
        r"""Return C code to extract a ``PyObject *`` instance.

        The code returned from this function must be templated using
        ``%(name)s``, representing the name that the caller wants to
        call this `Variable`. The Python object ``self.data`` is in a
        variable called ``"py_%(name)s"`` and this code must set the
        variables declared by :meth:`CLinkerType.c_declare` to something
        representative of ``py_%(name)``\s. If the data is improper, set an
        appropriate exception and insert ``"%(fail)s"``.

        TODO: Point out that template filling (via sub) is now performed
        by this function. --jpt

        Parameters
        ----------
        name
            The name of the ``PyObject *`` pointer that will store the value
            for this type.
        sub
            A dictionary of special codes. Most importantly
            ``sub['fail']``. See `CLinker` for more info on ``sub`` and
            ``fail``.

        Examples
        --------

        .. code-block: python

            def c_extract(self, name, sub, check_input=True, **kwargs):
                return "if (py_%(name)s == Py_None)" + \\\
                            addr_of_%(name)s = &py_%(name)s;" + \\\
                    "else" + \\\
                    { PyErr_SetString(PyExc_ValueError, \\\
                            'was expecting None'); %(fail)s;}"

        """

    @abstractmethod
    def c_sync(self, name: Text, sub: Dict[Text, Text]) -> Text:
        """Return C code to pack C types back into a ``PyObject``.

        The code returned from this function must be templated using
        ``"%(name)s"``, representing the name that the caller wants to
        call this `Variable`. The returned code may set ``"py_%(name)s"``
        to a ``PyObject*`` and that ``PyObject*`` will be accessible from
        Python via ``variable.data``. Do not forget to adjust reference
        counts if ``"py_%(name)s"`` is changed from its original value.

        Parameters
        ----------
        name
            WRITEME
        sub
            WRITEME

        """

    def c_element_type(self) -> Text:
        """Return the name of the primitive C type of items into variables handled by this type.

        e.g:

         - For ``TensorType(dtype='int64', ...)``: should return ``"npy_int64"``.
         - For ``GpuArrayType(dtype='int32', ...)``: should return ``"ga_int"``.

        """
        return ""

    def c_is_simple(self) -> bool:
        """Return ``True`` for small or builtin C types.

        A hint to tell the compiler that this type is a builtin C type or a
        small struct and that its memory footprint is negligible. Simple
        objects may be passed on the stack.

        """
        return False

    def c_literal(self, data: Constant) -> Text:
        """Provide a C literal string value for the specified `data`.

        Parameters
        ----------
        data
            The data to be converted into a C literal string.

        """
        return ""

    def c_extract_out(
        self, name: Text, sub: Dict[Text, Text], check_input: bool = True, **kwargs
    ) -> Text:
        """Return C code to extract a ``PyObject *`` instance.

        Unlike :math:`CLinkerType.c_extract`, :meth:`CLinkerType.c_extract_out` has to
        accept ``Py_None``, meaning that the variable should be left
        uninitialized.

        """
        return """
        if (py_%(name)s == Py_None)
        {
            %(c_init_code)s
        }
        else
        {
            %(c_extract_code)s
        }
        """ % dict(
            name=name,
            c_init_code=self.c_init(name, sub),
            c_extract_code=self.c_extract(name, sub, check_input),
        )

    def c_cleanup(self, name: Text, sub: Dict[Text, Text]) -> Text:
        """Return C code to clean up after :meth:`CLinkerType.c_extract`.

        This returns C code that should deallocate whatever
        :meth:`CLinkerType.c_extract` allocated or decrease the reference counts. Do
        not decrease ``py_%(name)s``'s reference count.

        Parameters
        ----------
        name : str
            WRITEME
        sub : dict of str
            WRITEME

        """
        return ""

    def c_code_cache_version(self) -> Union[Tuple, Tuple[int]]:
        """Return a tuple of integers indicating the version of this type.

        An empty tuple indicates an "unversioned" type that will not
        be cached between processes.

        The cache mechanism may erase cached modules that have been
        superseded by newer versions. See `ModuleCache` for details.

        """
        return ()


class HideC(CLinkerOp):
    def __hide(*args):
        raise MethodNotDefined()

    c_code: Callable = __hide
    c_code_cleanup: Callable = __hide

    c_headers: Callable = __hide
    c_header_dirs: Callable = __hide
    c_libraries: Callable = __hide
    c_lib_dirs: Callable = __hide

    c_support_code: Callable = __hide
    c_support_code_apply: Callable = __hide

    c_compile_args: Callable = __hide
    c_no_compile_args: Callable = __hide
    c_init_code: Callable = __hide
    c_init_code_apply: Callable = __hide

    c_init_code_struct: Callable = __hide
    c_support_code_struct: Callable = __hide
    c_cleanup_code_struct: Callable = __hide

    def c_code_cache_version(self):
        return ()

    def c_code_cache_version_apply(self, node):
        return self.c_code_cache_version()



import logging

import numpy as np

import aesara
from aesara import scalar as aes
from aesara.configdefaults import config
from aesara.graph.basic import Variable
from aesara.graph.type import CType
from aesara.misc.safe_asarray import _asarray
from aesara.utils import apply_across_args


_logger = logging.getLogger("aesara.tensor.type")


# Define common subsets of dtypes (as strings).
complex_dtypes = list(map(str, aes.complex_types))
continuous_dtypes = list(map(str, aes.continuous_types))
float_dtypes = list(map(str, aes.float_types))
integer_dtypes = list(map(str, aes.integer_types))
discrete_dtypes = list(map(str, aes.discrete_types))
all_dtypes = list(map(str, aes.all_types))
int_dtypes = list(map(str, aes.int_types))
uint_dtypes = list(map(str, aes.uint_types))

# TODO: add more type correspondences for e.g. int32, int64, float32,
# complex64, etc.
dtype_specs_map = {
    "float16": (float, "npy_float16", "NPY_FLOAT16"),
    "float32": (float, "npy_float32", "NPY_FLOAT32"),
    "float64": (float, "npy_float64", "NPY_FLOAT64"),
    "bool": (bool, "npy_bool", "NPY_BOOL"),
    "uint8": (int, "npy_uint8", "NPY_UINT8"),
    "int8": (int, "npy_int8", "NPY_INT8"),
    "uint16": (int, "npy_uint16", "NPY_UINT16"),
    "int16": (int, "npy_int16", "NPY_INT16"),
    "uint32": (int, "npy_uint32", "NPY_UINT32"),
    "int32": (int, "npy_int32", "NPY_INT32"),
    "uint64": (int, "npy_uint64", "NPY_UINT64"),
    "int64": (int, "npy_int64", "NPY_INT64"),
    "complex128": (complex, "aesara_complex128", "NPY_COMPLEX128"),
    "complex64": (complex, "aesara_complex64", "NPY_COMPLEX64"),
}


class TensorType(CType):
    """
    Symbolic `Type` representing a numpy.ndarray value.

    Initialize self.dtype and self.broadcastable.

    Parameters
    ----------
    dtype: str
        Corresponding to numpy dtype (e.g., 'int64')
        The value (ndarray) associated to a `Variable` of this `Type` will
        have this dtype.
    broadcastable: tuple, list, or array of boolean values
        This argument serves two purposes. First, the True elements of this
        list indicate the dimensions where the shape of an associated value
        must be 1. Secondly, the length of this list is the number of
        dimensions that an associated value must have. See
        doc:`broadcasting` for an explanation of how this list is used.
    name : str
        Optional name for this type.

    """

    __props__ = ("dtype", "broadcastable")

    context_name = "cpu"
    filter_checks_isfinite = False
    """
    When this is True, strict filtering rejects data containing NaN or
    Inf entries. (Used in `DebugMode`)
    """

    def __init__(self, dtype, broadcastable, name=None):
        if isinstance(dtype, str) and dtype == "floatX":
            self.dtype = config.floatX
        else:
            self.dtype = np.dtype(dtype).name

        # broadcastable is immutable, and all elements are either
        # True or False
        self.broadcastable = tuple(bool(b) for b in broadcastable)
        self.dtype_specs()  # error checking is done there
        self.name = name
        self.numpy_dtype = np.dtype(self.dtype)

    def clone(self, dtype=None, broadcastable=None):
        """
        Return a copy of the type optionally with a new dtype or
        broadcastable pattern.

        """
        if dtype is None:
            dtype = self.dtype
        if broadcastable is None:
            broadcastable = self.broadcastable
        return self.__class__(dtype, broadcastable, name=self.name)

    def filter(self, data, strict=False, allow_downcast=None):
        """
        Convert `data` to something which can be associated to a
        `TensorVariable`.

        This function is not meant to be called in user code. It is for
        `Linker` instances to use when running a compiled graph.

        """
        # Explicit error message when one accidentally uses a Variable as
        # input (typical mistake, especially with shared variables).
        if isinstance(data, Variable):
            raise TypeError(
                "Expected an array-like object, but found a Variable: "
                "maybe you are trying to call a function on a (possibly "
                "shared) variable instead of a numeric array?"
            )

        if isinstance(data, np.memmap) and (data.dtype == self.numpy_dtype):
            # numpy.memmap is a "safe" subclass of ndarray,
            # so we can use it wherever we expect a base ndarray.
            # however, casting it would defeat the purpose of not
            # loading the whole data into memory
            pass
        elif isinstance(data, np.ndarray) and (data.dtype == self.numpy_dtype):
            if data.dtype.num != self.numpy_dtype.num:
                data = _asarray(data, dtype=self.dtype)
            # -- now fall through to ndim check
        elif strict:
            # If any of the two conditions above was not met,
            # we raise a meaningful TypeError.
            if not isinstance(data, np.ndarray):
                raise TypeError(f"{self} expected a ndarray object (got {type(data)}).")
            if data.dtype != self.numpy_dtype:
                raise TypeError(
                    f"{self} expected an ndarray with dtype={self.numpy_dtype} (got {data.dtype})."
                )
        else:
            if allow_downcast:
                # Convert to self.dtype, regardless of the type of data
                data = _asarray(data, dtype=self.dtype)
                # TODO: consider to pad shape with ones to make it consistent
                # with self.broadcastable... like vector->row type thing
            else:
                if isinstance(data, np.ndarray):
                    # Check if self.dtype can accurately represent data
                    # (do not try to convert the data)
                    up_dtype = aes.upcast(self.dtype, data.dtype)
                    if up_dtype == self.dtype:
                        # Bug in the following line when data is a
                        # scalar array, see
                        # http://projects.scipy.org/numpy/ticket/1611
                        # data = data.astype(self.dtype)
                        data = _asarray(data, dtype=self.dtype)
                    if up_dtype != self.dtype:
                        err_msg = (
                            f"{self} cannot store a value of dtype {data.dtype} without "
                            "risking loss of precision. If you do not mind "
                            "this loss, you can: "
                            f"1) explicitly cast your data to {self.dtype}, or "
                            '2) set "allow_input_downcast=True" when calling '
                            f'"function". Value: "{repr(data)}"'
                        )
                        raise TypeError(err_msg)
                elif (
                    allow_downcast is None
                    and isinstance(data, (float, np.floating))
                    and self.dtype == config.floatX
                ):
                    # Special case where we allow downcasting of Python float
                    # literals to floatX, even when floatX=='float32'
                    data = _asarray(data, self.dtype)
                else:
                    # data has to be converted.
                    # Check that this conversion is lossless
                    converted_data = _asarray(data, self.dtype)
                    # We use the `values_eq` static function from TensorType
                    # to handle NaN values.
                    if TensorType.values_eq(
                        np.asarray(data), converted_data, force_same_dtype=False
                    ):
                        data = converted_data
                    else:
                        # Do not print a too long description of data
                        # (ndarray truncates it, but it's not sure for data)
                        str_data = str(data)
                        if len(str_data) > 80:
                            str_data = str_data[:75] + "(...)"

                        err_msg = (
                            f"{self} cannot store accurately value {data}, "
                            f"it would be represented as {converted_data}. "
                            "If you do not mind this precision loss, you can: "
                            "1) explicitly convert your data to a numpy array "
                            f"of dtype {self.dtype}, or "
                            '2) set "allow_input_downcast=True" when calling '
                            '"function".'
                        )
                        raise TypeError(err_msg)

        if self.ndim != data.ndim:
            raise TypeError(
                f"Wrong number of dimensions: expected {self.ndim},"
                f" got {data.ndim} with shape {data.shape}."
            )
        if not data.flags.aligned:
            raise TypeError(
                "The numpy.ndarray object is not aligned."
                " Aesara C code does not support that.",
            )

        i = 0
        for b in self.broadcastable:
            if b and data.shape[i] != 1:
                raise TypeError("Non-unit value on shape on a broadcastable dimension.")
            i += 1
        if self.filter_checks_isfinite and not np.all(np.isfinite(data)):
            raise ValueError("Non-finite elements not allowed")
        return data

    def filter_variable(self, other, allow_convert=True):
        """
        Convert a symbolic Variable into a TensorType, if compatible.

        For the moment, only a TensorType and GpuArrayType will be
        converted, provided they have the same number of dimensions
        and dtype and have "compatible" broadcastable pattern.

        """
        if not isinstance(other, Variable):
            # The value is not a Variable: we cast it into
            # a Constant of the appropriate Type.
            other = self.Constant(type=self, data=other)

        if other.type == self:
            return other

        if allow_convert:
            # Attempt safe broadcast conversion.
            other2 = self.convert_variable(other)
            if other2 is not None and other2.type == self:
                return other2

        raise TypeError(
            f"Cannot convert Type {other.type} "
            f"(of Variable {other}) into Type {self}. "
            f"You can try to manually convert {other} into a {self}."
        )

    def value_validity_msg(self, a):
        try:
            self.filter(a, strict=True)
        except Exception as e:
            return str(e)
        return "value is valid"

    def dtype_specs(self):
        """
        Return a tuple (python type, c type, numpy typenum) that corresponds
        to self.dtype.

        This function is used internally as part of C code generation.

        """
        try:
            return dtype_specs_map[self.dtype]
        except KeyError:
            raise TypeError(
                f"Unsupported dtype for {self.__class__.__name__}: {self.dtype}"
            )

    def to_scalar_type(self):
        return aes.get_scalar_type(dtype=self.dtype)

    def __eq__(self, other):
        """
        Compare True iff other is the same kind of TensorType.

        """
        return (
            type(self) == type(other)
            and other.dtype == self.dtype
            and other.broadcastable == self.broadcastable
        )

    def convert_variable(self, var):
        if (
            isinstance(self, type(var.type))
            and self.dtype == var.type.dtype  # noqa
            and self.ndim == var.type.ndim
            and all(
                sb == ob or ob
                for sb, ob in zip(self.broadcastable, var.type.broadcastable)
            )
        ):
            return aesara.tensor.patternbroadcast(var, self.broadcastable)

    @staticmethod
    def may_share_memory(a, b):
        # This is a method of TensorType, so both a and b should be ndarrays
        if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
            return np.may_share_memory(a, b)
        else:
            return False

    @staticmethod
    def values_eq(a, b, force_same_dtype=True):
        # TODO: check to see if the shapes must match
        #      for now, we err on safe side...
        if a.shape != b.shape:
            return False
        if force_same_dtype and a.dtype != b.dtype:
            return False
        a_eq_b = a == b
        r = np.all(a_eq_b)
        if r:
            return True
        # maybe the trouble is that there are NaNs
        a_missing = np.isnan(a)
        if a_missing.any():
            b_missing = np.isnan(b)
            return np.all(a_eq_b + (a_missing == b_missing))
        else:
            return False

    @staticmethod
    def values_eq_approx(
        a, b, allow_remove_inf=False, allow_remove_nan=False, rtol=None, atol=None
    ):
        return values_eq_approx(a, b, allow_remove_inf, allow_remove_nan, rtol, atol)

    def __hash__(self):
        return hash((type(self), self.dtype, self.broadcastable))

    ndim = property(lambda self: len(self.broadcastable), doc="number of dimensions")
    """
    Number of dimensions.

    This read-only property is the preferred way to get the number of
    dimensions of a `TensorType`.

    """

    def make_variable(self, name=None):
        """
        Return a `TensorVariable` of this type.

        Parameters
        ----------
        name : str
            A pretty name to identify this `Variable` when printing and
            debugging

        """
        return self.Variable(self, name=name)

    def __str__(self):
        if self.name:
            return self.name
        else:
            b = self.broadcastable
            named_broadcastable = {
                (): "scalar",
                (False,): "vector",
                (False, True): "col",
                (True, False): "row",
                (False, False): "matrix",
            }
            if b in named_broadcastable:
                bcast = named_broadcastable[b]
            else:
                if any(b):
                    bcast = str(b)
                else:
                    bcast = f"{len(b)}D"
            return f"TensorType({self.dtype}, {bcast})"

    def __repr__(self):
        return str(self)

    def c_element_type(self):
        return self.dtype_specs()[1]

    def c_declare(self, name, sub, check_input=True):
        if check_input:
            check = """
            typedef %(dtype)s dtype_%(name)s;
            """ % dict(
                sub, name=name, dtype=self.dtype_specs()[1]
            )
        else:
            check = ""
        declaration = """
        PyArrayObject* %(name)s;
        """ % dict(
            sub, name=name, dtype=self.dtype_specs()[1]
        )

        return declaration + check

    def c_init(self, name, sub):
        return """
        %(name)s = NULL;
        """ % dict(
            sub, name=name, type_num=self.dtype_specs()[2]
        )

    def c_extract(self, name, sub, check_input=True, **kwargs):
        if check_input:
            check = """
            %(name)s = NULL;
            if (py_%(name)s == Py_None) {
                // We can either fail here or set %(name)s to NULL and rely on Ops
                // using tensors to handle the NULL case, but if they fail to do so
                // they'll end up with nasty segfaults, so this is public service.
                PyErr_SetString(PyExc_ValueError, "expected an ndarray, not None");
                %(fail)s
            }
            if (!PyArray_Check(py_%(name)s)) {
                PyErr_SetString(PyExc_ValueError, "expected an ndarray");
                %(fail)s
            }
            // We expect %(type_num)s
            if (!PyArray_ISALIGNED((PyArrayObject*) py_%(name)s)) {
                PyArrayObject * tmp = (PyArrayObject*) py_%(name)s;
                PyErr_Format(PyExc_NotImplementedError,
                             "expected an aligned array of type %%ld "
                             "(%(type_num)s), got non-aligned array of type %%ld"
                             " with %%ld dimensions, with 3 last dims "
                             "%%ld, %%ld, %%ld"
                             " and 3 last strides %%ld %%ld, %%ld.",
                             (long int) %(type_num)s,
                             (long int) PyArray_TYPE((PyArrayObject*) py_%(name)s),
                             (long int) PyArray_NDIM(tmp),
                             (long int) (PyArray_NDIM(tmp) >= 3 ?
            PyArray_DIMS(tmp)[PyArray_NDIM(tmp)-3] : -1),
                             (long int) (PyArray_NDIM(tmp) >= 2 ?
            PyArray_DIMS(tmp)[PyArray_NDIM(tmp)-2] : -1),
                             (long int) (PyArray_NDIM(tmp) >= 1 ?
            PyArray_DIMS(tmp)[PyArray_NDIM(tmp)-1] : -1),
                             (long int) (PyArray_NDIM(tmp) >= 3 ?
            PyArray_STRIDES(tmp)[PyArray_NDIM(tmp)-3] : -1),
                             (long int) (PyArray_NDIM(tmp) >= 2 ?
            PyArray_STRIDES(tmp)[PyArray_NDIM(tmp)-2] : -1),
                             (long int) (PyArray_NDIM(tmp) >= 1 ?
            PyArray_STRIDES(tmp)[PyArray_NDIM(tmp)-1] : -1)
            );
                %(fail)s
            }
            // This is a TypeError to be consistent with DEBUG_MODE
            // Note: DEBUG_MODE also tells the name of the container
            if (PyArray_TYPE((PyArrayObject*) py_%(name)s) != %(type_num)s) {
                PyErr_Format(PyExc_TypeError,
                             "expected type_num %%d (%(type_num)s) got %%d",
                             %(type_num)s, PyArray_TYPE((PyArrayObject*) py_%(name)s));
                %(fail)s
            }
            """ % dict(
                sub, name=name, type_num=self.dtype_specs()[2]
            )
        else:
            check = ""
        return (
            check
            + """
        %(name)s = (PyArrayObject*)(py_%(name)s);
        Py_XINCREF(%(name)s);
        """
            % dict(sub, name=name, type_num=self.dtype_specs()[2])
        )

    def c_cleanup(self, name, sub):
        return (
            """
        if (%(name)s) {
            Py_XDECREF(%(name)s);
        }
        """
            % locals()
        )

    def c_sync(self, name, sub):
        fail = sub["fail"]
        type_num = self.dtype_specs()[2]
        return (
            """
        {Py_XDECREF(py_%(name)s);}
        if (!%(name)s) {
            Py_INCREF(Py_None);
            py_%(name)s = Py_None;
        }
        else if ((void*)py_%(name)s != (void*)%(name)s) {
            py_%(name)s = (PyObject*)%(name)s;
        }

        {Py_XINCREF(py_%(name)s);}

        if (%(name)s && !PyArray_ISALIGNED((PyArrayObject*) py_%(name)s)) {
            PyErr_Format(PyExc_NotImplementedError,
                         "c_sync: expected an aligned array, got non-aligned array of type %%ld"
                         " with %%ld dimensions, with 3 last dims "
                         "%%ld, %%ld, %%ld"
                         " and 3 last strides %%ld %%ld, %%ld.",
                         (long int) PyArray_TYPE((PyArrayObject*) py_%(name)s),
                         (long int) PyArray_NDIM(%(name)s),
                         (long int) (PyArray_NDIM(%(name)s) >= 3 ?
        PyArray_DIMS(%(name)s)[PyArray_NDIM(%(name)s)-3] : -1),
                         (long int) (PyArray_NDIM(%(name)s) >= 2 ?
        PyArray_DIMS(%(name)s)[PyArray_NDIM(%(name)s)-2] : -1),
                         (long int) (PyArray_NDIM(%(name)s) >= 1 ?
        PyArray_DIMS(%(name)s)[PyArray_NDIM(%(name)s)-1] : -1),
                         (long int) (PyArray_NDIM(%(name)s) >= 3 ?
        PyArray_STRIDES(%(name)s)[PyArray_NDIM(%(name)s)-3] : -1),
                         (long int) (PyArray_NDIM(%(name)s) >= 2 ?
        PyArray_STRIDES(%(name)s)[PyArray_NDIM(%(name)s)-2] : -1),
                         (long int) (PyArray_NDIM(%(name)s) >= 1 ?
        PyArray_STRIDES(%(name)s)[PyArray_NDIM(%(name)s)-1] : -1)
        );
            %(fail)s
        }
        """
            % locals()
        )

    def c_headers(self, **kwargs):
        return aes.get_scalar_type(self.dtype).c_headers(**kwargs)

    def c_libraries(self, **kwargs):
        return aes.get_scalar_type(self.dtype).c_libraries(**kwargs)

    def c_compile_args(self, **kwargs):
        return aes.get_scalar_type(self.dtype).c_compile_args(**kwargs)

    def c_support_code(self, **kwargs):
        return aes.get_scalar_type(self.dtype).c_support_code(**kwargs)

    def c_init_code(self, **kwargs):
        return aes.get_scalar_type(self.dtype).c_init_code(**kwargs)

    def c_code_cache_version(self):
        scalar_version = aes.get_scalar_type(self.dtype).c_code_cache_version()
        if scalar_version:
            return (11,) + scalar_version
        else:
            return ()

    def value_zeros(self, shape):
        """Create an numpy ndarray full of 0 values."""
        return np.zeros(shape, dtype=self.dtype)

    def get_shape_info(self, obj):
        """Return the information needed to compute the memory size of `obj`.

        The memory size is only the data, so this excludes the container.
        For an ndarray, this is the data, but not the ndarray object and
        other data structures such as shape and strides.

        `get_shape_info` and `get_size` work in tandem for the memory
        profiler.

        `get_shape_info` is called during the execution of the function.
        So it is better that it is not too slow.

        `get_size` will be called on the output of this function
        when printing the memory profile.

        Parameters
        ----------
        obj
            The object that this Type represents during execution.

        Returns
        -------
        object
            Python object that can be passed to `get_size`.

        """
        return obj.shape

    def get_size(self, shape_info):
        """Number of bytes taken by the object represented by `shape_info`.

        Parameters
        ----------
        shape_info
            The output of the call to `get_shape_info`.

        Returns
        -------
        int
            The number of bytes taken by the object described by ``shape_info``.

        """
        if shape_info:
            return np.prod(shape_info) * np.dtype(self.dtype).itemsize
        else:  # a scalar
            return np.dtype(self.dtype).itemsize


def values_eq_approx(
    a, b, allow_remove_inf=False, allow_remove_nan=False, rtol=None, atol=None
):
    """
    Parameters
    ----------
    allow_remove_inf
        If True, when there is an inf in a, we allow any value in b in
        that position. Event -inf
    allow_remove_nan
        If True, when there is a nan in a, we allow any value in b in
        that position. Event +-inf
    rtol
        Relative tolerance, passed to _allclose.
    atol
        Absolute tolerance, passed to _allclose.

    """
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        if a.shape != b.shape:
            return False
        if a.dtype != b.dtype:
            return False
        if str(a.dtype) not in continuous_dtypes:
            return np.all(a == b)
        else:
            cmp = aesara.tensor.math._allclose(a, b, rtol=rtol, atol=atol)
            if cmp:
                # Numpy claims they are close, this is good enough for us.
                return True
            # Numpy is unhappy, but it does not necessarily mean that a and
            # b are different. Indeed, Numpy does not like missing values
            # and will return False whenever some are found in a or b.
            # The proper way would be to use the MaskArray stuff available
            # in Numpy. However, it looks like it has been added to Numpy's
            # core recently, so it may not be available to everyone. Thus,
            # for now we use a home-made recipe, that should probably be
            # revisited in the future.
            a_missing = np.isnan(a)
            a_inf = np.isinf(a)

            if not (a_missing.any() or (allow_remove_inf and a_inf.any())):
                # There are no missing values in a, thus this is not the
                # reason why numpy.allclose(a, b) returned False.
                _logger.info(
                    f"numpy allclose failed for abs_err {np.max(abs(a - b))} and rel_err {np.max(abs(a - b) / (abs(a) + abs(b)))}"
                )
                return False
            # The following line is what numpy.allclose bases its decision
            # upon, according to its documentation.
            rtol = 1.0000000000000001e-05
            atol = 1e-8
            cmp_elemwise = np.absolute(a - b) <= (atol + rtol * np.absolute(b))
            # Find places where both a and b have missing values.
            both_missing = a_missing * np.isnan(b)

            # Find places where both a and b have inf of the same sign.
            both_inf = a_inf * np.isinf(b)

            # cmp_elemwise is weird when we have inf and -inf.
            # set it to False
            cmp_elemwise = np.where(both_inf & cmp_elemwise, a == b, cmp_elemwise)

            # check the sign of the inf
            both_inf = np.where(both_inf, (a == b), both_inf)

            if allow_remove_inf:
                both_inf += a_inf
            if allow_remove_nan:
                both_missing += a_missing

            # Combine all information.
            return (cmp_elemwise + both_missing + both_inf).all()

    return False


def values_eq_approx_remove_inf(a, b):
    return values_eq_approx(a, b, True)


def values_eq_approx_remove_nan(a, b):
    return values_eq_approx(a, b, False, True)


def values_eq_approx_remove_inf_nan(a, b):
    return values_eq_approx(a, b, True, True)


def values_eq_approx_always_true(a, b):
    return True


aesara.compile.register_view_op_c_code(
    TensorType,
    """
    Py_XDECREF(%(oname)s);
    %(oname)s = %(iname)s;
    Py_XINCREF(%(oname)s);
    """,
    version=1,
)


aesara.compile.register_deep_copy_op_c_code(
    TensorType,
    """
    int alloc = %(oname)s == NULL;
    for(int i=0; !alloc && i<PyArray_NDIM(%(oname)s); i++) {
       if(PyArray_DIMS(%(iname)s)[i] != PyArray_DIMS(%(oname)s)[i]) {
           alloc = true;
           break;
       }
    }
    if(alloc) {
        Py_XDECREF(%(oname)s);
        %(oname)s = (PyArrayObject*)PyArray_NewCopy(%(iname)s,
                                                    NPY_ANYORDER);
        if (!%(oname)s)
        {
            PyErr_SetString(PyExc_ValueError,
                            "DeepCopyOp: the copy failed!");
            %(fail)s;
        }
    } else {
        if(PyArray_CopyInto(%(oname)s, %(iname)s)){
            PyErr_SetString(PyExc_ValueError,
        "DeepCopyOp: the copy failed into already allocated space!");
            %(fail)s;
        }
    }
    """,
    version=2,
)


def tensor(*args, **kwargs):
    name = kwargs.pop("name", None)
    return TensorType(*args, **kwargs)(name=name)


cscalar = TensorType("complex64", ())
zscalar = TensorType("complex128", ())
fscalar = TensorType("float32", ())
dscalar = TensorType("float64", ())
bscalar = TensorType("int8", ())
wscalar = TensorType("int16", ())
iscalar = TensorType("int32", ())
lscalar = TensorType("int64", ())


def scalar(name=None, dtype=None):
    """Return a symbolic scalar variable.

    Parameters
    ----------
    dtype: numeric
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, ())
    return type(name)


scalars, fscalars, dscalars, iscalars, lscalars = apply_across_args(
    scalar, fscalar, dscalar, iscalar, lscalar
)

int_types = bscalar, wscalar, iscalar, lscalar
float_types = fscalar, dscalar
complex_types = cscalar, zscalar
int_scalar_types = int_types
float_scalar_types = float_types
complex_scalar_types = complex_types

cvector = TensorType("complex64", (False,))
zvector = TensorType("complex128", (False,))
fvector = TensorType("float32", (False,))
dvector = TensorType("float64", (False,))
bvector = TensorType("int8", (False,))
wvector = TensorType("int16", (False,))
ivector = TensorType("int32", (False,))
lvector = TensorType("int64", (False,))


def vector(name=None, dtype=None):
    """Return a symbolic vector variable.

    Parameters
    ----------
    dtype: numeric
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False,))
    return type(name)


vectors, fvectors, dvectors, ivectors, lvectors = apply_across_args(
    vector, fvector, dvector, ivector, lvector
)

int_vector_types = bvector, wvector, ivector, lvector
float_vector_types = fvector, dvector
complex_vector_types = cvector, zvector

cmatrix = TensorType("complex64", (False, False))
zmatrix = TensorType("complex128", (False, False))
fmatrix = TensorType("float32", (False, False))
dmatrix = TensorType("float64", (False, False))
bmatrix = TensorType("int8", (False, False))
wmatrix = TensorType("int16", (False, False))
imatrix = TensorType("int32", (False, False))
lmatrix = TensorType("int64", (False, False))


def matrix(name=None, dtype=None):
    """Return a symbolic matrix variable.

    Parameters
    ----------
    dtype: numeric
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False, False))
    return type(name)


matrices, fmatrices, dmatrices, imatrices, lmatrices = apply_across_args(
    matrix, fmatrix, dmatrix, imatrix, lmatrix
)

int_matrix_types = bmatrix, wmatrix, imatrix, lmatrix
float_matrix_types = fmatrix, dmatrix
complex_matrix_types = cmatrix, zmatrix

crow = TensorType("complex64", (True, False))
zrow = TensorType("complex128", (True, False))
frow = TensorType("float32", (True, False))
drow = TensorType("float64", (True, False))
brow = TensorType("int8", (True, False))
wrow = TensorType("int16", (True, False))
irow = TensorType("int32", (True, False))
lrow = TensorType("int64", (True, False))


def row(name=None, dtype=None):
    """Return a symbolic row variable (ndim=2, broadcastable=[True,False]).

    Parameters
    ----------
    dtype: numeric type
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (True, False))
    return type(name)


rows, frows, drows, irows, lrows = apply_across_args(row, frow, drow, irow, lrow)

ccol = TensorType("complex64", (False, True))
zcol = TensorType("complex128", (False, True))
fcol = TensorType("float32", (False, True))
dcol = TensorType("float64", (False, True))
bcol = TensorType("int8", (False, True))
wcol = TensorType("int16", (False, True))
icol = TensorType("int32", (False, True))
lcol = TensorType("int64", (False, True))


def col(name=None, dtype=None):
    """Return a symbolic column variable (ndim=2, broadcastable=[False,True]).

    Parameters
    ----------
    dtype : numeric
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False, True))
    return type(name)


cols, fcols, dcols, icols, lcols = apply_across_args(col, fcol, dcol, icol, lcol)

ctensor3 = TensorType("complex64", ((False,) * 3))
ztensor3 = TensorType("complex128", ((False,) * 3))
ftensor3 = TensorType("float32", ((False,) * 3))
dtensor3 = TensorType("float64", ((False,) * 3))
btensor3 = TensorType("int8", ((False,) * 3))
wtensor3 = TensorType("int16", ((False,) * 3))
itensor3 = TensorType("int32", ((False,) * 3))
ltensor3 = TensorType("int64", ((False,) * 3))


def tensor3(name=None, dtype=None):
    """Return a symbolic 3-D variable.

    Parameters
    ----------
    dtype: numeric type
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False, False, False))
    return type(name)


tensor3s, ftensor3s, dtensor3s, itensor3s, ltensor3s = apply_across_args(
    tensor3, ftensor3, dtensor3, itensor3, ltensor3
)

ctensor4 = TensorType("complex64", ((False,) * 4))
ztensor4 = TensorType("complex128", ((False,) * 4))
ftensor4 = TensorType("float32", ((False,) * 4))
dtensor4 = TensorType("float64", ((False,) * 4))
btensor4 = TensorType("int8", ((False,) * 4))
wtensor4 = TensorType("int16", ((False,) * 4))
itensor4 = TensorType("int32", ((False,) * 4))
ltensor4 = TensorType("int64", ((False,) * 4))


def tensor4(name=None, dtype=None):
    """Return a symbolic 4-D variable.

    Parameters
    ----------
    dtype: numeric type
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False, False, False, False))
    return type(name)


tensor4s, ftensor4s, dtensor4s, itensor4s, ltensor4s = apply_across_args(
    tensor4, ftensor4, dtensor4, itensor4, ltensor4
)

ctensor5 = TensorType("complex64", ((False,) * 5))
ztensor5 = TensorType("complex128", ((False,) * 5))
ftensor5 = TensorType("float32", ((False,) * 5))
dtensor5 = TensorType("float64", ((False,) * 5))
btensor5 = TensorType("int8", ((False,) * 5))
wtensor5 = TensorType("int16", ((False,) * 5))
itensor5 = TensorType("int32", ((False,) * 5))
ltensor5 = TensorType("int64", ((False,) * 5))


def tensor5(name=None, dtype=None):
    """Return a symbolic 5-D variable.

    Parameters
    ----------
    dtype: numeric type
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False, False, False, False, False))
    return type(name)


tensor5s, ftensor5s, dtensor5s, itensor5s, ltensor5s = apply_across_args(
    tensor5, ftensor5, dtensor5, itensor5, ltensor5
)

ctensor6 = TensorType("complex64", ((False,) * 6))
ztensor6 = TensorType("complex128", ((False,) * 6))
ftensor6 = TensorType("float32", ((False,) * 6))
dtensor6 = TensorType("float64", ((False,) * 6))
btensor6 = TensorType("int8", ((False,) * 6))
wtensor6 = TensorType("int16", ((False,) * 6))
itensor6 = TensorType("int32", ((False,) * 6))
ltensor6 = TensorType("int64", ((False,) * 6))


def tensor6(name=None, dtype=None):
    """Return a symbolic 6-D variable.

    Parameters
    ----------
    dtype: numeric type
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False,) * 6)
    return type(name)


tensor6s, ftensor6s, dtensor6s, itensor6s, ltensor6s = apply_across_args(
    tensor6, ftensor6, dtensor6, itensor6, ltensor6
)

ctensor7 = TensorType("complex64", ((False,) * 7))
ztensor7 = TensorType("complex128", ((False,) * 7))
ftensor7 = TensorType("float32", ((False,) * 7))
dtensor7 = TensorType("float64", ((False,) * 7))
btensor7 = TensorType("int8", ((False,) * 7))
wtensor7 = TensorType("int16", ((False,) * 7))
itensor7 = TensorType("int32", ((False,) * 7))
ltensor7 = TensorType("int64", ((False,) * 7))


def tensor7(name=None, dtype=None):
    """Return a symbolic 7-D variable.

    Parameters
    ----------
    dtype: numeric type
        None means to use aesara.config.floatX.
    name
        A name to attach to this variable.

    """
    if dtype is None:
        dtype = config.floatX
    type = TensorType(dtype, (False,) * 7)
    return type(name)


tensor7s, ftensor7s, dtensor7s, itensor7s, ltensor7s = apply_across_args(
    tensor7, ftensor7, dtensor7, itensor7, ltensor7
)


__all__ = [
    "TensorType",
    "bcol",
    "bmatrix",
    "brow",
    "bscalar",
    "btensor3",
    "btensor4",
    "btensor5",
    "btensor6",
    "btensor7",
    "bvector",
    "ccol",
    "cmatrix",
    "col",
    "cols",
    "complex_matrix_types",
    "complex_scalar_types",
    "complex_types",
    "complex_vector_types",
    "crow",
    "cscalar",
    "ctensor3",
    "ctensor4",
    "ctensor5",
    "ctensor6",
    "ctensor7",
    "cvector",
    "dcol",
    "dcols",
    "dmatrices",
    "dmatrix",
    "drow",
    "drows",
    "dscalar",
    "dscalars",
    "dtensor3",
    "dtensor3s",
    "dtensor4",
    "dtensor4s",
    "dtensor5",
    "dtensor5s",
    "dtensor6",
    "dtensor6s",
    "dtensor7",
    "dtensor7s",
    "dvector",
    "dvectors",
    "fcol",
    "fcols",
    "float_matrix_types",
    "float_scalar_types",
    "float_types",
    "float_vector_types",
    "fmatrices",
    "fmatrix",
    "frow",
    "frows",
    "fscalar",
    "fscalars",
    "ftensor3",
    "ftensor3s",
    "ftensor4",
    "ftensor4s",
    "ftensor5",
    "ftensor5s",
    "ftensor6",
    "ftensor6s",
    "ftensor7",
    "ftensor7s",
    "fvector",
    "fvectors",
    "icol",
    "icols",
    "imatrices",
    "imatrix",
    "int_matrix_types",
    "int_scalar_types",
    "int_types",
    "int_vector_types",
    "irow",
    "irows",
    "iscalar",
    "iscalars",
    "itensor3",
    "itensor3s",
    "itensor4",
    "itensor4s",
    "itensor5",
    "itensor5s",
    "itensor6",
    "itensor6s",
    "itensor7",
    "itensor7s",
    "ivector",
    "ivectors",
    "lcol",
    "lcols",
    "lmatrices",
    "lmatrix",
    "lrow",
    "lrows",
    "lscalar",
    "lscalars",
    "ltensor3",
    "ltensor3s",
    "ltensor4",
    "ltensor4s",
    "ltensor5",
    "ltensor5s",
    "ltensor6",
    "ltensor6s",
    "ltensor7",
    "ltensor7s",
    "lvector",
    "lvectors",
    "matrices",
    "matrix",
    "row",
    "rows",
    "scalar",
    "scalars",
    "tensor",
    "tensor3",
    "tensor3s",
    "tensor4",
    "tensor4s",
    "tensor5",
    "tensor5s",
    "tensor6",
    "tensor6s",
    "tensor7",
    "tensor7s",
    "values_eq_approx_always_true",
    "vector",
    "vectors",
    "wcol",
    "wrow",
    "wscalar",
    "wtensor3",
    "wtensor4",
    "wtensor5",
    "wtensor6",
    "wtensor7",
    "wvector",
    "zcol",
    "zmatrix",
    "zrow",
    "zscalar",
    "ztensor3",
    "ztensor4",
    "ztensor5",
    "ztensor6",
    "ztensor7",
    "zvector",
]



import copy

import scipy.sparse

from aesara.compile import SharedVariable, shared_constructor
from aesara.sparse.basic import SparseType, _sparse_py_operators


class SparseTensorSharedVariable(_sparse_py_operators, SharedVariable):
    dtype = property(lambda self: self.type.dtype)
    format = property(lambda self: self.type.format)


@shared_constructor
def sparse_constructor(
    value, name=None, strict=False, allow_downcast=None, borrow=False, format=None
):
    """
    SharedVariable Constructor for SparseType.

    writeme

    """
    if not isinstance(value, scipy.sparse.spmatrix):
        raise TypeError(
            "Expected a sparse matrix in the sparse shared variable constructor. Received: ",
            value.__class__,
        )

    if format is None:
        format = value.format
    type = SparseType(format=format, dtype=value.dtype)
    if not borrow:
        value = copy.deepcopy(value)
    return SparseTensorSharedVariable(
        type=type, value=value, name=name, strict=strict, allow_downcast=allow_downcast
    )



import warnings


warnings.warn(
    "The module `aesara.assert_op` is deprecated "
    "and its `Op`s have been moved to `aesara.raise_op`",
    DeprecationWarning,
    stacklevel=2,
)

from aesara.raise_op import Assert, assert_op  # noqa: F401 E402



"""
Classes and functions for validating graphs that contain view
and inplace operations.

"""
import itertools
from collections import OrderedDict, deque

import aesara
from aesara.configdefaults import config
from aesara.graph.basic import Constant
from aesara.graph.features import AlreadyThere, Bookkeeper
from aesara.graph.fg import InconsistencyError
from aesara.misc.ordered_set import OrderedSet


class ProtocolError(Exception):
    """
    Raised when FunctionGraph calls DestroyHandler callbacks in
    an invalid way, for example, pruning or changing a node that has
    never been imported.

    """


def _contains_cycle(fgraph, orderings):
    """
    Function to check if the given graph contains a cycle

    Parameters
    ----------
    fgraph
        The FunctionGraph to check for cycles.
    orderings
        Dictionary specifying extra dependencies besides those encoded in
        Variable.owner / Apply.inputs.

        If orderings[my_apply] == dependencies, then my_apply is an Apply
        instance, dependencies is a set of Apply instances, and every member
        of dependencies must be executed before my_apply.

        The dependencies are typically used to prevent
        inplace apply nodes from destroying their input before
        other apply nodes with the same input access it.

    Returns
    -------
    bool
        True if the graph contains a cycle, False otherwise.

    """
    # These are lists of Variable instances
    outputs = fgraph.outputs

    # this is hard-coded reimplementation of functions from graph.py
    # reason: go faster, prepare for port to C.
    # specifically, it could be replaced with a wrapper
    # around graph.io_toposort that returns True iff io_toposort raises
    # a ValueError containing the substring 'cycle'.
    # This implementation is optimized for the destroyhandler and runs
    # slightly faster than io_toposort.

    # this is performance-critical code. it is the largest single-function
    # bottleneck when compiling large graphs.
    assert isinstance(outputs, (tuple, list, deque))

    # TODO: For more speed - use a defaultdict for the orderings
    # (defaultdict runs faster than dict in the case where the key
    # is not in the dictionary, at least in CPython)

    # IG: I tried converting parent_counts to use an id for the key,
    # so that the dict would do reference counting on its keys.
    # This caused a slowdown.
    # Separate benchmark tests showed that calling id is about
    # half as expensive as a dictionary access, and that the
    # dictionary also runs slower when storing ids than when
    # storing objects.

    # dict mapping an Apply or Variable instance to the number
    # of its parents (including parents imposed by orderings)
    # that haven't been visited yet
    parent_counts = {}
    # dict mapping an Apply or Variable instance to its children
    node_to_children = {}

    # visitable: A container holding all Variable and Apply instances
    # that can currently be visited according to the graph topology
    # (ie, whose parents have already been visited)
    # TODO: visitable is a fifo_queue. could this run faster if we
    # implement it as a stack rather than a deque?
    # TODO: visitable need not be a fifo_queue, any kind of container
    # that we can throw things into and take things out of quickly will
    # work. is there another kind of container that could run faster?
    # we don't care about the traversal order here as much as we do
    # in io_toposort because we aren't trying to generate an ordering
    # on the nodes
    visitable = deque()

    # IG: visitable could in principle be initialized to fgraph.inputs
    #     + fgraph.orphans... if there were an fgraph.orphans structure.
    #     I tried making one and maintaining it caused a huge slowdown.
    #     This may be because I made it a list, so it would have a
    #     deterministic iteration order, in hopes of using it to speed
    #     up toposort as well.
    #     I think since we need to scan through all variables and nodes
    #     to make parent_counts anyway, it's cheap enough to always
    #     detect orphans at cycle detection / toposort time

    # Pass through all the nodes to build visitable, parent_count, and
    # node_to_children
    for var in fgraph.variables:

        # this is faster than calling get_parents
        owner = var.owner
        # variables don't appear in orderings, so we don't need to worry
        # about that here
        if owner:
            # insert node in node_to_children[r]
            # (if r is not already in node_to_children,
            # initialize it to [])
            node_to_children.setdefault(owner, []).append(var)
            parent_counts[var] = 1
        else:
            visitable.append(var)
            parent_counts[var] = 0

    for a_n in fgraph.apply_nodes:
        parents = list(a_n.inputs)
        # This is faster than conditionally extending
        # IG: I tried using a shared empty_list = [] constructed
        # outside of the for loop to avoid constructing multiple
        # lists, but this was not any faster.
        parents.extend(orderings.get(a_n, []))

        if parents:
            for parent in parents:
                # insert node in node_to_children[r]
                # (if r is not already in node_to_children,
                # initialize it to [])
                node_to_children.setdefault(parent, []).append(a_n)
            parent_counts[a_n] = len(parents)
        else:
            # an Apply with no inputs would be a weird case, but I'm
            # not sure we forbid it
            visitable.append(a_n)
            parent_counts[a_n] = 0

    # at this point,
    # parent_counts.keys() == fgraph.apply_nodes + fgraph.variables

    # Now we actually check for cycles
    # As long as there are nodes that can be visited while respecting
    # the topology, we keep visiting nodes
    # If we run out of visitable nodes and we haven't visited all nodes,
    # then there was a cycle. It blocked the traversal because some
    # node couldn't be visited until one of its descendants had been
    # visited too.
    # This is a standard cycle detection algorithm.

    visited = 0
    while visitable:
        # Since each node is inserted into the visitable queue exactly
        # once, it comes out of the queue exactly once
        # That means we can decrement its children's unvisited parent count
        # and increment the visited node count without double-counting
        node = visitable.popleft()
        visited += 1
        for client in node_to_children.get(node, []):
            parent_counts[client] -= 1
            # If all of a node's parents have been visited,
            # it may now be visited too
            if not parent_counts[client]:
                visitable.append(client)

    return visited != len(parent_counts)


def _build_droot_impact(destroy_handler):
    droot = {}  # destroyed view + nonview variables -> foundation
    impact = {}  # destroyed nonview variable -> it + all views of it
    root_destroyer = {}  # root -> destroyer apply

    for app in destroy_handler.destroyers:
        for output_idx, input_idx_list in app.op.destroy_map.items():
            if len(input_idx_list) != 1:
                raise NotImplementedError()
            input_idx = input_idx_list[0]
            input = app.inputs[input_idx]

            # Find non-view variable which is ultimately viewed by input.
            view_i = destroy_handler.view_i
            _r = input
            while _r is not None:
                r = _r
                _r = view_i.get(r)
            input_root = r

            if input_root in droot:
                raise InconsistencyError(f"Multiple destroyers of {input_root}")
            droot[input_root] = input_root
            root_destroyer[input_root] = app

            # The code here add all the variables that are views of r into
            # an OrderedSet input_impact
            input_impact = OrderedSet()

            q = deque()
            q.append(input_root)
            while len(q) > 0:
                v = q.popleft()
                for n in destroy_handler.view_o.get(v, []):
                    input_impact.add(n)
                    q.append(n)

            for v in input_impact:
                assert v not in droot
                droot[v] = input_root

            impact[input_root] = input_impact
            impact[input_root].add(input_root)

    return droot, impact, root_destroyer


def fast_inplace_check(fgraph, inputs):
    """
    Return the variables in inputs that are possible candidate for as inputs of
    inplace operation.

    Parameters
    ----------
    inputs : list
        Inputs Variable that you want to use as inplace destination.

    """
    Supervisor = aesara.compile.function.types.Supervisor
    protected_inputs = [
        f.protected for f in fgraph._features if isinstance(f, Supervisor)
    ]
    protected_inputs = sum(protected_inputs, [])  # flatten the list
    protected_inputs.extend(fgraph.outputs)

    inputs = [
        i
        for i in inputs
        if not isinstance(i, Constant)
        and not fgraph.has_destroyers([i])
        and i not in protected_inputs
    ]
    return inputs


class DestroyHandler(Bookkeeper):  # noqa
    """
    The DestroyHandler class detects when a graph is impossible to evaluate
    because of aliasing and destructive operations.

    Several data structures are used to do this.

    An Op can use its view_map property to declare that an output may be
    aliased to an input. If that output is destroyed, the input is also
    considered to be destroyed. The view_maps of several Ops can feed into
    one another and form a directed graph. The consequence of destroying any
    variable in such a graph is that all variables in the graph must be
    considered to be destroyed, because they could all be referring to the
    same underlying storage.

    In the current implementation, that graph is a tree, and the root of that
    tree is called the foundation.

    TODO: why "in the current implementation" ? is there another implementation
          planned?
    TODO: why is the graph a tree? isn't it possible that one variable could
          be aliased to many variables? for example, don't switch and ifelse
          have to do this?

    The original DestroyHandler (if 0'ed out above) computed several data
    structures from scratch each time it was asked to validate the graph.
    Because this happens potentially thousands of times and each graph to
    validate is extremely similar to the previous one, computing the
    data structures from scratch repeatedly was wasteful and resulted in
    high compile times for large graphs.

    This implementation computes the data structures once at initialization
    and then incrementally updates them.

    It is a work in progress. The following data structures have been
    converted to use the incremental strategy:
        <none>

    The following data structures remain to be converted:
        <unknown>

    """

    pickle_rm_attr = ["destroyers", "has_destroyers"]

    def __init__(self, do_imports_on_attach=True, algo=None):
        self.fgraph = None
        self.do_imports_on_attach = do_imports_on_attach

        """
        Maps every variable in the graph to its "foundation" (deepest
        ancestor in view chain).
        TODO: change name to var_to_vroot.

        """
        self.droot = OrderedDict()

        """
        Maps a variable to all variables that are indirect or direct views of it
        (including itself) essentially the inverse of droot.
        TODO: do all variables appear in this dict, or only those that are
              foundations?
        TODO: do only destroyed variables go in here? one old docstring said so.
        TODO: rename to x_to_views after reverse engineering what x is

        """
        self.impact = OrderedDict()

        """
        If a var is destroyed, then this dict will map
        droot[var] to the apply node that destroyed var
        TODO: rename to vroot_to_destroyer

        """
        self.root_destroyer = OrderedDict()
        if algo is None:
            algo = config.cycle_detection
        self.algo = algo
        self.fail_validate = OrderedDict()

    def on_attach(self, fgraph):
        """
        When attaching to a new fgraph, check that
            1) This DestroyHandler wasn't already attached to some fgraph
               (its data structures are only set up to serve one).
            2) The FunctionGraph doesn't already have a DestroyHandler.
               This would result in it validating everything twice, causing
               compilation to be slower.

        Give the FunctionGraph instance:
            1) A new method "destroyers(var)"
               TODO: what does this do exactly?
            2) A new attribute, "destroy_handler"
        TODO: WRITEME: what does this do besides the checks?

        """

        # Do the checking #
        already_there = False
        if self.fgraph is fgraph:
            already_there = True
        if self.fgraph is not None:
            raise Exception(
                "A DestroyHandler instance can only serve one"
                " FunctionGraph. (Matthew 6:24)"
            )
        for attr in ("destroyers", "destroy_handler"):
            if hasattr(fgraph, attr):
                already_there = True

        if already_there:
            # FunctionGraph.attach_feature catches AlreadyThere and cancels the attachment
            raise AlreadyThere(
                "DestroyHandler feature is already present"
                " or in conflict with another plugin."
            )

        # Annotate the FunctionGraph #
        self.unpickle(fgraph)
        fgraph.destroy_handler = self

        self.fgraph = fgraph
        self.destroyers = (
            OrderedSet()
        )  # set of Apply instances with non-null destroy_map
        self.view_i = {}  # variable -> variable used in calculation
        self.view_o = (
            {}
        )  # variable -> set of variables that use this one as a direct input
        # clients: how many times does an apply use a given variable
        self.clients = OrderedDict()  # variable -> apply -> ninputs
        self.stale_droot = True

        self.debug_all_apps = set()
        if self.do_imports_on_attach:
            Bookkeeper.on_attach(self, fgraph)

    def unpickle(self, fgraph):
        def get_destroyers_of(r):
            droot, _, root_destroyer = self.refresh_droot_impact()
            try:
                return [root_destroyer[droot[r]]]
            except Exception:
                return []

        fgraph.destroyers = get_destroyers_of

        def has_destroyers(protected_list):
            if self.algo != "fast":
                droot, _, root_destroyer = self.refresh_droot_impact()
                for protected_var in protected_list:
                    try:
                        root_destroyer[droot[protected_var]]
                        return True
                    except KeyError:
                        pass
                return False

            def recursive_destroys_finder(protected_var):
                # protected_var is the idx'th input of app.
                for (app, idx) in fgraph.clients[protected_var]:
                    if app == "output":
                        continue
                    destroy_maps = app.op.destroy_map.values()
                    # If True means that the apply node, destroys the protected_var.
                    if idx in [dmap for sublist in destroy_maps for dmap in sublist]:
                        return True
                    for var_idx in app.op.view_map.keys():
                        if idx in app.op.view_map[var_idx]:
                            # We need to recursively check the destroy_map of all the
                            # outputs that we have a view_map on.
                            if recursive_destroys_finder(app.outputs[var_idx]):
                                return True
                return False

            for protected_var in protected_list:
                if recursive_destroys_finder(protected_var):
                    return True
            return False

        fgraph.has_destroyers = has_destroyers

    def refresh_droot_impact(self):
        """
        Makes sure self.droot, self.impact, and self.root_destroyer are up to
        date, and returns them (see docstrings for these properties above).

        """
        if self.stale_droot:
            self.droot, self.impact, self.root_destroyer = _build_droot_impact(self)
            self.stale_droot = False
        return self.droot, self.impact, self.root_destroyer

    def on_detach(self, fgraph):
        if fgraph is not self.fgraph:
            raise Exception("detaching wrong fgraph", fgraph)
        del self.destroyers
        del self.view_i
        del self.view_o
        del self.clients
        del self.stale_droot
        assert self.fgraph.destroyer_handler is self
        delattr(self.fgraph, "destroyers")
        delattr(self.fgraph, "has_destroyers")
        delattr(self.fgraph, "destroy_handler")
        self.fgraph = None

    def fast_destroy(self, fgraph, app, reason):
        """
        Do the check for only 1 level.

        For now:
        - Destroyed variables can have only 1 clients.
        - Allow view to have multiple clients.
        - Allow sequence of view.
        - But don't allow to destroy view
        """
        dm = app.op.destroy_map
        if not dm:
            return
        inputs = set(
            itertools.chain.from_iterable(dm.values())
        )  # list of app's destroyed inputs
        for inp_idx in inputs:
            inp = app.inputs[inp_idx]
            if getattr(inp.tag, "indestructible", False) or isinstance(inp, Constant):
                self.fail_validate[app] = InconsistencyError(
                    f"Attempting to destroy indestructible variables: {inp}"
                )
            elif len(fgraph.clients[inp]) > 1:
                self.fail_validate[app] = InconsistencyError(
                    "Destroyed variable has more than one client. " + str(reason)
                )
            elif inp.owner:
                app2 = inp.owner
                inp_idx2 = app2.outputs.index(inp)
                v = app2.op.view_map
                d = app2.op.destroy_map
                if v:
                    v = v.get(inp_idx2, [])
                    if len(v) > 0:
                        self.fail_validate[app] = InconsistencyError(
                            "Destroyed variable has view_map. " + str(reason)
                        )
                elif d:
                    d = d.get(inp_idx2, [])
                    if len(d) > 0:
                        self.fail_validate[app] = InconsistencyError(
                            "Destroyed variable has destroy_map. " + str(reason)
                        )

                # These 2 assertions are commented since this function is called so many times
                # but they should be true.
                # assert len(v) <= 1
                # assert len(d) <= 1

    def on_import(self, fgraph, app, reason):
        """
        Add Apply instance to set which must be computed.

        """
        if app in self.debug_all_apps:
            raise ProtocolError("double import")
        self.debug_all_apps.add(app)
        # print 'DH IMPORT', app, id(app), id(self), len(self.debug_all_apps)

        # If it's a destructive op, add it to our watch list
        dmap = app.op.destroy_map
        vmap = app.op.view_map
        if dmap:
            self.destroyers.add(app)
            if self.algo == "fast":
                self.fast_destroy(fgraph, app, reason)

        # add this symbol to the forward and backward maps
        for o_idx, i_idx_list in vmap.items():
            if len(i_idx_list) > 1:
                raise NotImplementedError(
                    "destroying this output invalidates multiple inputs", (app.op)
                )
            o = app.outputs[o_idx]
            i = app.inputs[i_idx_list[0]]
            self.view_i[o] = i
            self.view_o.setdefault(i, OrderedSet()).add(o)

        # update self.clients
        for i, input in enumerate(app.inputs):
            self.clients.setdefault(input, OrderedDict()).setdefault(app, 0)
            self.clients[input][app] += 1

        for i, output in enumerate(app.outputs):
            self.clients.setdefault(output, OrderedDict())

        self.stale_droot = True

    def on_prune(self, fgraph, app, reason):
        """
        Remove Apply instance from set which must be computed.

        """
        if app not in self.debug_all_apps:
            raise ProtocolError("prune without import")
        self.debug_all_apps.remove(app)

        # UPDATE self.clients
        for input in set(app.inputs):
            del self.clients[input][app]

        if app.op.destroy_map:
            self.destroyers.remove(app)

        # Note: leaving empty client dictionaries in the struct.
        # Why? It's a pain to remove them. I think they aren't doing any harm, they will be
        # deleted on_detach().

        # UPDATE self.view_i, self.view_o
        for o_idx, i_idx_list in app.op.view_map.items():
            if len(i_idx_list) > 1:
                # destroying this output invalidates multiple inputs
                raise NotImplementedError()
            o = app.outputs[o_idx]
            i = app.inputs[i_idx_list[0]]

            del self.view_i[o]

            self.view_o[i].remove(o)
            if not self.view_o[i]:
                del self.view_o[i]

        self.stale_droot = True
        if app in self.fail_validate:
            del self.fail_validate[app]

    def on_change_input(self, fgraph, app, i, old_r, new_r, reason):
        """
        app.inputs[i] changed from old_r to new_r.

        """
        if app == "output":
            # app == 'output' is special key that means FunctionGraph is redefining which nodes are being
            # considered 'outputs' of the graph.
            pass
        else:
            if app not in self.debug_all_apps:
                raise ProtocolError("change without import")

            # UPDATE self.clients
            self.clients[old_r][app] -= 1
            if self.clients[old_r][app] == 0:
                del self.clients[old_r][app]

            self.clients.setdefault(new_r, OrderedDict()).setdefault(app, 0)
            self.clients[new_r][app] += 1

            # UPDATE self.view_i, self.view_o
            for o_idx, i_idx_list in app.op.view_map.items():
                if len(i_idx_list) > 1:
                    # destroying this output invalidates multiple inputs
                    raise NotImplementedError()
                i_idx = i_idx_list[0]
                output = app.outputs[o_idx]
                if i_idx == i:
                    if app.inputs[i_idx] is not new_r:
                        raise ProtocolError("wrong new_r on change")

                    self.view_i[output] = new_r

                    self.view_o[old_r].remove(output)
                    if not self.view_o[old_r]:
                        del self.view_o[old_r]

                    self.view_o.setdefault(new_r, OrderedSet()).add(output)

            if self.algo == "fast":
                if app in self.fail_validate:
                    del self.fail_validate[app]
                self.fast_destroy(fgraph, app, reason)
        self.stale_droot = True

    def validate(self, fgraph):
        """
        Return None.

        Raise InconsistencyError when
        a) orderings() raises an error
        b) orderings cannot be topologically sorted.

        """
        if self.destroyers:
            if self.algo == "fast":
                if self.fail_validate:
                    app_err_pairs = self.fail_validate
                    self.fail_validate = OrderedDict()
                    # self.fail_validate can only be a hint that maybe/probably
                    # there is a cycle.This is because inside replace() we could
                    # record many reasons to not accept a change, but we don't
                    # know which one will fail first inside validate(). Thus,the
                    # graph might have already changed when we raise the
                    # self.fail_validate error. So before raising the error, we
                    # double check here.
                    for app in app_err_pairs:
                        if app in fgraph.apply_nodes:
                            self.fast_destroy(fgraph, app, "validate")
                    if self.fail_validate:
                        self.fail_validate = app_err_pairs
                        raise app_err_pairs[app]
            else:
                ords = self.orderings(fgraph, ordered=False)
                if _contains_cycle(fgraph, ords):
                    raise InconsistencyError("Dependency graph contains cycles")
        else:
            # James's Conjecture:
            # If there are no destructive ops, then there can be no cycles.

            # FB: This isn't always True. It can happened that
            # optimization introduce node that depend on itself. This
            # is very rare and should not happen in general. It will be
            # caught later. The error will be far from the source. But
            # doing this conjecture should speed up compilation most of
            # the time. The user should create such dependency except
            # if he mess too much with the internal.
            pass
        return True

    def orderings(self, fgraph, ordered=True):
        """
        Return orderings induced by destructive operations.

        Raise InconsistencyError when
        a) attempting to destroy indestructable variable, or
        b) attempting to destroy a value multiple times, or
        c) an Apply destroys (illegally) one of its own inputs by aliasing

        """
        if ordered:
            set_type = OrderedSet
            rval = OrderedDict()
        else:
            set_type = set
            rval = dict()

        if self.destroyers:
            # BUILD DATA STRUCTURES
            # CHECK for multiple destructions during construction of variables

            droot, impact, __ignore = self.refresh_droot_impact()

            # check for destruction of constants
            illegal_destroy = [
                r
                for r in droot
                if getattr(r.tag, "indestructible", False) or isinstance(r, Constant)
            ]
            if illegal_destroy:
                raise InconsistencyError(
                    f"Attempting to destroy indestructible variables: {illegal_destroy}"
                )

            # add destroyed variable clients as computational dependencies
            for app in self.destroyers:
                # keep track of clients that should run before the current Apply
                root_clients = set_type()
                # for each destroyed input...
                for output_idx, input_idx_list in app.op.destroy_map.items():
                    destroyed_idx = input_idx_list[0]
                    destroyed_variable = app.inputs[destroyed_idx]
                    root = droot[destroyed_variable]
                    root_impact = impact[root]
                    # we generally want to put all clients of things which depend on root
                    # as pre-requisites of app.
                    # But, app is itself one such client!
                    # App will always be a client of the node we're destroying
                    # (destroyed_variable, but the tricky thing is when it is also a client of
                    # *another variable* viewing on the root.  Generally this is illegal, (e.g.,
                    # add_inplace(x, x.T).  In some special cases though, the in-place op will
                    # actually be able to work properly with multiple destroyed inputs (e.g,
                    # add_inplace(x, x).  An Op that can still work in this case should declare
                    # so via the 'destroyhandler_tolerate_same' attribute or
                    # 'destroyhandler_tolerate_aliased' attribute.
                    #
                    # destroyhandler_tolerate_same should be a list of pairs of the form
                    # [(idx0, idx1), (idx0, idx2), ...]
                    # The first element of each pair is the input index of a destroyed
                    # variable.
                    # The second element of each pair is the index of a different input where
                    # we will permit exactly the same variable to appear.
                    # For example, add_inplace.tolerate_same might be [(0,1)] if the destroyed
                    # input is also allowed to appear as the second argument.
                    #
                    # destroyhandler_tolerate_aliased is the same sort of list of
                    # pairs.
                    # op.destroyhandler_tolerate_aliased = [(idx0, idx1)] tells the
                    # destroyhandler to IGNORE an aliasing between a destroyed
                    # input idx0 and another input idx1.
                    # This is generally a bad idea, but it is safe in some
                    # cases, such as
                    # - the op reads from the aliased idx1 before modifying idx0
                    # - the idx0 and idx1 are guaranteed not to overlap (e.g.
                    #   they are pointed at different rows of a matrix).
                    #

                    # CHECK FOR INPUT ALIASING
                    # OPT: pre-compute this on import
                    tolerate_same = getattr(app.op, "destroyhandler_tolerate_same", [])
                    assert isinstance(tolerate_same, list)
                    tolerated = {
                        idx1 for idx0, idx1 in tolerate_same if idx0 == destroyed_idx
                    }
                    tolerated.add(destroyed_idx)
                    tolerate_aliased = getattr(
                        app.op, "destroyhandler_tolerate_aliased", []
                    )
                    assert isinstance(tolerate_aliased, list)
                    ignored = {
                        idx1 for idx0, idx1 in tolerate_aliased if idx0 == destroyed_idx
                    }
                    for i, input in enumerate(app.inputs):
                        if i in ignored:
                            continue
                        if input in root_impact and (
                            i not in tolerated or input is not destroyed_variable
                        ):
                            raise InconsistencyError(
                                f"Input aliasing: {app} ({destroyed_idx}, {i})"
                            )

                    # add the rule: app must be preceded by all other Apply instances that
                    # depend on destroyed_input
                    for r in root_impact:
                        assert not [a for a, c in self.clients[r].items() if not c]
                        root_clients.update(
                            [a for a, c in self.clients[r].items() if c]
                        )

                # app itself is a client of the destroyed inputs,
                # but should not run before itself
                root_clients.remove(app)
                if root_clients:
                    rval[app] = root_clients

        return rval



from warnings import warn


try:
    import scipy

    enable_sparse = True
except ImportError:
    enable_sparse = False
    warn("SciPy can't be imported.  Sparse matrix support is disabled.")

from aesara.sparse.type import SparseType, _is_sparse


if enable_sparse:
    from aesara.sparse import opt, sharedvar
    from aesara.sparse.basic import *
    from aesara.sparse.sharedvar import sparse_constructor as shared

    def sparse_grad(var):
        """This function return a new variable whose gradient will be
        stored in a sparse format instead of dense.

        Currently only variable created by AdvancedSubtensor1 is supported.
        i.e. a_tensor_var[an_int_vector].

        .. versionadded:: 0.6rc4
        """
        from aesara.tensor.subtensor import AdvancedSubtensor, AdvancedSubtensor1

        if var.owner is None or not isinstance(
            var.owner.op, (AdvancedSubtensor, AdvancedSubtensor1)
        ):
            raise TypeError(
                "Sparse gradient is only implemented for AdvancedSubtensor and AdvancedSubtensor1"
            )

        x = var.owner.inputs[0]
        indices = var.owner.inputs[1:]

        if len(indices) > 1:
            raise TypeError(
                "Sparse gradient is only implemented for single advanced indexing"
            )

        ret = AdvancedSubtensor1(sparse_grad=True)(x, indices[0])
        return ret



import copy
import traceback as tb
import warnings
from collections.abc import Iterable
from numbers import Number
from typing import Optional

import numpy as np

from aesara import tensor as at
from aesara.configdefaults import config
from aesara.graph.basic import Constant, Variable
from aesara.scalar import ComplexError, IntegerDivisionError
from aesara.tensor.exceptions import AdvancedIndexingError
from aesara.tensor.type import TensorType
from aesara.tensor.utils import hash_from_ndarray


class _tensor_py_operators:
    def __abs__(self):
        return at.math.abs(self)

    def __neg__(self):
        return at.math.neg(self)

    # These won't work because Python requires an int return value
    # def __int__(self): return convert_to_int32(self)
    # def __float__(self): return convert_to_float64(self)
    # def __complex__(self): return convert_to_complex128(self)

    _is_nonzero = True

    def __lt__(self, other):
        rval = at.math.lt(self, other)
        rval._is_nonzero = False
        return rval

    def __le__(self, other):
        rval = at.math.le(self, other)
        rval._is_nonzero = False
        return rval

    def __gt__(self, other):
        rval = at.math.gt(self, other)
        rval._is_nonzero = False
        return rval

    def __ge__(self, other):
        rval = at.math.ge(self, other)
        rval._is_nonzero = False
        return rval

    def __bool__(self):
        # This is meant to prohibit stuff like a < b < c, which is internally
        # implemented as (a < b) and (b < c). The trouble with this is the
        # side-effect that checking for a non-NULL a by typing "if a: ..."
        # uses the same __nonzero__ method.  We want these both to work, but
        # it seems impossible.  Currently, all vars evaluate to nonzero except
        # the return values of comparison operators, which raise this
        # exception.  If you can think of a better solution, go for it!
        #
        # __bool__ is Python 3.x data model. __nonzero__ is Python 2.x.
        if self._is_nonzero:
            return True
        else:
            raise TypeError("Variables do not support boolean operations.")

    def __invert__(self):
        return at.math.invert(self)

    def __and__(self, other):
        return at.math.and_(self, other)

    def __or__(self, other):
        return at.math.or_(self, other)

    def __xor__(self, other):
        return at.math.xor(self, other)

    def __rand__(self, other):
        return at.math.and_(other, self)

    def __ror__(self, other):
        return at.math.or_(other, self)

    def __rxor__(self, other):
        return at.math.xor(other, self)

    # def __iand__(self, other):
    #    return _and_inplace(self, other)
    #
    # def __ior__(self, other):
    #    return _or_inplace(self, other)
    #
    # def __ixor__(self, other):
    #    return _xor_inplace(self, other)

    def __add__(self, other):
        try:
            return at.math.add(self, other)
        # We should catch the minimum number of exception here.
        # Otherwise this will convert error when Aesara flags
        # compute_test_value is used
        # Evidently, we need to catch NotImplementedError
        # TypeError from as_tensor_variable are caught in Elemwise.make_node
        # Otherwise TensorVariable * SparseVariable won't work!
        except (NotImplementedError, TypeError):
            # We must return NotImplemented and not an
            # NotImplementedError or raise an NotImplementedError.
            # That way python will give a good error message like this
            # `TypeError: unsupported operand type(s) for +:
            # 'TensorVariable' and 'TensorVariable'`
            return NotImplemented

    def __sub__(self, other):
        # See explanation in __add__ for the error caught
        # and the return value in that case
        try:
            return at.math.sub(self, other)
        except (NotImplementedError, TypeError):
            return NotImplemented

    def __mul__(self, other):
        # See explanation in __add__ for the error caught
        # and the return value in that case
        try:
            return at.math.mul(self, other)
        except (NotImplementedError, TypeError):
            return NotImplemented

    def __div__(self, other):
        # See explanation in __add__ for the error caught
        # and the return value in that case
        try:
            return at.math.div_proxy(self, other)
        except IntegerDivisionError:
            # This is to raise the exception that occurs when trying to divide
            # two integer arrays (currently forbidden).
            raise
        except (NotImplementedError, TypeError):
            return NotImplemented

    def __pow__(self, other):
        # See explanation in __add__ for the error caught
        # and the return value in that case
        try:
            return at.math.pow(self, other)
        except (NotImplementedError, TypeError):
            return NotImplemented

    def __mod__(self, other):
        # See explanation in __add__ for the error caught
        # and the return value in that case
        try:
            return at.math.mod_check(self, other)
        except ComplexError:
            # This is to raise the exception that occurs when trying to compute
            # x % y with either x or y a complex number.
            raise
        except (NotImplementedError, TypeError):
            return NotImplemented

    def __divmod__(self, other):
        return at.math.divmod(self, other)

    def __truediv__(self, other):
        return at.math.true_div(self, other)

    def __floordiv__(self, other):
        return at.math.floor_div(self, other)

    def __rtruediv__(self, other):
        return at.math.true_div(other, self)

    def __rfloordiv__(self, other):
        return at.math.floor_div(other, self)

    # Do not use these; in-place `Op`s should be inserted by optimizations
    # only!
    # def __iadd__(self, other):
    #    return _add_inplace(self, other)
    # def __isub__(self, other):
    #    return _sub_inplace(self, other)
    #
    # def __imul__(self, other):
    #    return _mul_inplace(self, other)
    #
    # def __idiv__(self, other):
    #    return _div_inplace(self, other)
    #
    # def __ipow__(self, other):
    #    return _pow_inplace(self, other)

    def __radd__(self, other):
        return at.math.add(other, self)

    def __rsub__(self, other):
        return at.math.sub(other, self)

    def __rmul__(self, other):
        return at.math.mul(other, self)

    def __rdiv__(self, other):
        return at.math.div_proxy(other, self)

    def __rmod__(self, other):
        return at.math.mod(other, self)

    def __rdivmod__(self, other):
        return at.math.divmod(other, self)

    def __rpow__(self, other):
        return at.math.pow(other, self)

    def __ceil__(self):
        return at.math.ceil(self)

    def __floor__(self):
        return at.math.floor(self)

    def __trunc__(self):
        return at.math.trunc(self)

    # NumPy-like transpose property
    @property
    def T(self):
        return at.basic.transpose(self)

    def transpose(self, *axes):
        """Transpose this array.

        Returns
        -------
        object
            `tensor.transpose(self, axes)` or `tensor.transpose(self, axes[0])`.

        If only one `axes` argument is provided and it is iterable, then it is
        assumed to be the entire axes tuple, and passed intact to
        tensor.transpose.

        """
        if len(axes) == 0:
            return at.basic.transpose(self)
        try:
            iter(axes[0])
            iterable = True
        except TypeError:
            iterable = False
        if len(axes) == 1 and iterable:
            return at.basic.transpose(self, axes[0])
        else:
            return at.basic.transpose(self, axes)

    @property
    def shape(self):
        return at.shape(self)

    @property
    def size(self):
        if self.ndim == 1:
            return self.shape[0]
        else:
            return at.math.prod(self.shape)

    def any(self, axis=None, keepdims=False):
        return at.math.any(self, axis=axis, keepdims=keepdims)

    def all(self, axis=None, keepdims=False):
        return at.math.all(self, axis=axis, keepdims=keepdims)

    # Old note: "We can't implement this because Python requests that this
    # function returns an integer."
    # TODO: We could use `get_vector_length` and let it raise an exception just like
    # `__iter__` does
    # def __len__(self):
    #     raise Exception("Aesara Variables can't work with len(Aesara "
    #                     "Variable) due to Python restriction. You can use "
    #                     "AesaraVariable.shape[0] instead.")

    def reshape(self, shape, ndim=None):
        """Return a reshaped view/copy of this variable.

        Parameters
        ----------
        shape
            Something that can be converted to a symbolic vector of integers.
        ndim
            The length of the shape. Passing None here means for
            Aesara to try and guess the length of `shape`.


        .. warning:: This has a different signature than numpy's
                     ndarray.reshape!
                     In numpy you do not need to wrap the shape arguments
                     in a tuple, in aesara you do need to.

        """
        if ndim is not None:
            if not isinstance(ndim, int):
                raise ValueError(
                    "Expected ndim to be an integer, is " + str(type(ndim))
                )

        return at.reshape(self, shape, ndim=ndim)

    def dimshuffle(self, *pattern):
        """
        Reorder the dimensions of this variable, optionally inserting
        broadcasted dimensions.

        Parameters
        ----------
        pattern
            List/tuple of int mixed with 'x' for broadcastable dimensions.

        Examples
        --------
        For example, to create a 3D view of a [2D] matrix, call
        ``dimshuffle([0,'x',1])``.  This will create a 3D view such that the
        middle dimension is an implicit broadcasted dimension.  To do the same
        thing on the transpose of that matrix, call ``dimshuffle([1, 'x', 0])``.

        Notes
        -----
        This function supports the pattern passed as a tuple, or as a
        variable-length argument (e.g. ``a.dimshuffle(pattern)`` is equivalent
        to ``a.dimshuffle(*pattern)`` where ``pattern`` is a list/tuple of ints
        mixed with 'x' characters).

        See Also
        --------
        DimShuffle

        """
        if (len(pattern) == 1) and (isinstance(pattern[0], (list, tuple))):
            pattern = pattern[0]
        op = at.elemwise.DimShuffle(list(self.type.broadcastable), pattern)
        return op(self)

    def flatten(self, ndim=1):
        return at.basic.flatten(self, ndim)

    def ravel(self):
        return at.basic.flatten(self)

    def diagonal(self, offset=0, axis1=0, axis2=1):
        return at.basic.diagonal(self, offset, axis1, axis2)

    def transfer(self, target):
        """Transfer this this array's data to another device.

        If `target` is `'cpu'` this will transfer to a TensorType (if
        not already one).  Other types may define additional targets.

        Parameters
        ----------
        target : str
            The desired location of the output variable
        """
        return at.basic.transfer(self, target)

    def arccos(self):
        return at.math.arccos(self)

    def arccosh(self):
        return at.math.arccosh(self)

    def arcsin(self):
        return at.math.arcsin(self)

    def arcsinh(self):
        return at.math.arcsinh(self)

    def arctan(self):
        return at.math.arctan(self)

    def arctanh(self):
        return at.math.arctanh(self)

    def ceil(self):
        return at.math.ceil(self)

    def cos(self):
        return at.math.cos(self)

    def cosh(self):
        return at.math.cosh(self)

    def deg2rad(self):
        return at.math.deg2rad(self)

    def exp(self):
        return at.math.exp(self)

    def exp2(self):
        return at.math.exp2(self)

    def expm1(self):
        return at.math.expm1(self)

    def floor(self):
        return at.math.floor(self)

    def log(self):
        return at.math.log(self)

    def log10(self):
        return at.math.log10(self)

    def log1p(self):
        return at.math.log1p(self)

    def log2(self):
        return at.math.log2(self)

    def rad2deg(self):
        return at.math.rad2deg(self)

    def sin(self):
        return at.math.sin(self)

    def sinh(self):
        return at.math.sinh(self)

    def sqrt(self):
        return at.math.sqrt(self)

    def tan(self):
        return at.math.tan(self)

    def tanh(self):
        return at.math.tanh(self)

    def trunc(self):
        return at.math.trunc(self)

    def astype(self, dtype):
        return at.basic.cast(self, dtype)

    def __getitem__(self, args):
        def includes_bool(args_el):
            if isinstance(args_el, (np.bool_, bool)) or (
                hasattr(args_el, "dtype") and args_el.dtype == "bool"
            ):
                return True
            if not isinstance(args_el, Variable) and isinstance(args_el, Iterable):
                for el in args_el:
                    if includes_bool(el):
                        return True
            return False

        if isinstance(args, list) and any(isinstance(a, slice) for a in args):
            pass
        elif not isinstance(args, tuple):
            args = (args,)

        # Count the dimensions, check for bools and find ellipses.
        ellipses = []
        index_dim_count = 0
        for i, arg in enumerate(args):
            if arg is np.newaxis:
                # no increase in index_dim_count
                pass
            elif arg is Ellipsis:
                # no increase in index_dim_count
                ellipses.append(i)
            elif (
                isinstance(arg, (np.ndarray, Variable))
                and hasattr(arg, "dtype")
                and arg.dtype == "bool"
            ):
                index_dim_count += arg.ndim
            else:
                # Python arrays can contain a mixture of bools and integers,
                # which requires complex rules to handle all special cases.
                # These rules differ slightly between NumPy versions.
                # Since earlier versions of Aesara did not support any boolean
                # indexing, it is safe to throw an error if we encounter
                # any of these difficult cases.
                if includes_bool(arg):
                    raise TypeError(
                        "TensorType does not support Python bools "
                        "for indexing, such as tensor[[True, False]]. "
                        "To use a boolean mask, convert the mask to "
                        "a NumPy array first, e.g., "
                        "tensor[numpy.array([True, False])]."
                    )
                index_dim_count += 1

        # Check if the number of dimensions isn't too large.
        if self.ndim < index_dim_count:
            raise IndexError("too many indices for array")

        # Convert an Ellipsis if provided into an appropriate number of
        # slice(None).
        if len(ellipses) > 1:
            raise IndexError("an index can only have a single Ellipsis (`...`)")
        elif len(ellipses) == 1:
            ellipsis_at = ellipses[0]
            args = list(args)
            args[ellipsis_at : ellipsis_at + 1] = [slice(None)] * (
                self.ndim - index_dim_count
            )

        def is_empty_array(val):
            return (isinstance(val, (tuple, list)) and len(val) == 0) or (
                isinstance(val, np.ndarray) and val.size == 0
            )

        # Force input to be int64 datatype if input is an empty list or tuple
        # Else leave it as is if it is a real number
        # Convert python literals to aesara constants
        args = tuple(
            [
                at.subtensor.as_index_constant(
                    np.array(inp, dtype=np.int64) if is_empty_array(inp) else inp
                )
                for inp in args
            ]
        )

        # Determine if advanced indexing is needed or not.  The logic is
        # already in `index_vars_to_types`: if it succeeds, standard indexing is
        # used; if it fails with `AdvancedIndexingError`, advanced indexing is
        # used
        advanced = False
        for i, arg in enumerate(args):
            if includes_bool(arg):
                advanced = True
                break

            if arg is not np.newaxis:
                try:
                    at.subtensor.index_vars_to_types(arg)
                except AdvancedIndexingError:
                    if advanced:
                        break
                    else:
                        advanced = True

        if advanced:
            return at.subtensor.advanced_subtensor(self, *args)
        else:
            if np.newaxis in args:
                # `np.newaxis` (i.e. `None`) in NumPy indexing mean "add a new
                # broadcastable dimension at this location".  Since Aesara adds
                # new broadcastable dimensions via the `DimShuffle` `Op`, the
                # following code uses said `Op` to add one of the new axes and
                # then uses recursion to apply any other indices and add any
                # remaining new axes.

                counter = 0
                pattern = []
                new_args = []
                for arg in args:
                    if arg == np.newaxis:
                        pattern.append("x")
                        new_args.append(slice(None, None, None))
                    else:
                        pattern.append(counter)
                        counter += 1
                        new_args.append(arg)

                pattern.extend(list(range(counter, self.ndim)))

                view = self.dimshuffle(pattern)
                full_slices = True
                for arg in new_args:
                    # We can't do arg == slice(None, None, None) as in
                    # Python 2.7, this call __lt__ if we have a slice
                    # with some symbolic variable.
                    if not (
                        isinstance(arg, slice)
                        and arg.start is None
                        and arg.stop is None
                        and arg.step is None
                    ):
                        full_slices = False
                if full_slices:
                    return view
                else:
                    return view.__getitem__(tuple(new_args))
            else:
                return at.subtensor.Subtensor(args)(
                    self,
                    *at.subtensor.get_slice_elements(
                        args, lambda entry: isinstance(entry, Variable)
                    ),
                )

    def take(self, indices, axis=None, mode="raise"):
        return at.subtensor.take(self, indices, axis, mode)

    def copy(self, name=None):
        """Return a symbolic copy and optionally assign a name.

        Does not copy the tags.
        """
        copied_variable = at.basic.tensor_copy(self)
        copied_variable.name = name
        return copied_variable

    def __iter__(self):
        try:
            for i in range(at.basic.get_vector_length(self)):
                yield self[i]
        except TypeError:
            # This prevents accidental iteration via sum(self)
            raise TypeError(
                "TensorType does not support iteration. "
                "Maybe you are using builtins.sum instead of "
                "aesara.tensor.math.sum? (Maybe .max?)"
            )

    @property
    def ndim(self):
        """The rank of this tensor."""
        return self.type.ndim

    @property
    def broadcastable(self):
        """
        The broadcastable signature of this tensor.

        See Also
        --------
        broadcasting

        """
        return self.type.broadcastable

    @property
    def dtype(self):
        """The dtype of this tensor."""
        return self.type.dtype

    def __dot__(left, right):
        return at.math.dense_dot(left, right)

    def __rdot__(right, left):
        return at.math.dense_dot(left, right)

    dot = __dot__
    __matmul__ = __dot__
    __rmatmul__ = __rdot__

    def sum(self, axis=None, dtype=None, keepdims=False, acc_dtype=None):
        """See `aesara.tensor.math.sum`."""
        return at.math.sum(
            self, axis=axis, dtype=dtype, keepdims=keepdims, acc_dtype=acc_dtype
        )

    def prod(self, axis=None, dtype=None, keepdims=False, acc_dtype=None):
        """See `aesara.tensor.math.prod`."""
        return at.math.prod(
            self, axis=axis, dtype=dtype, keepdims=keepdims, acc_dtype=acc_dtype
        )

    def norm(self, L, axis=None, keepdims=False):
        if L == 0:
            raise NotImplementedError()
        if np.isinf(L):
            raise NotImplementedError()
        # optimizations will/should catch cases like L=1, L=2
        y = at.math.pow(
            at.math.pow(at.math.abs(self), L).sum(axis=axis),
            1.0 / L,
        )
        if keepdims:
            return at.math.makeKeepDims(self, y, axis)
        else:
            return y

    def mean(self, axis=None, dtype=None, keepdims=False, acc_dtype=None):
        """See `aesara.tensor.math.mean`."""
        return at.math.mean(
            self, axis=axis, dtype=dtype, keepdims=keepdims, acc_dtype=acc_dtype
        )

    def var(self, axis=None, ddof=0, keepdims=False, corrected=False):
        """See `aesara.tensor.math.var`."""
        return at.math.var(
            self, axis=axis, ddof=ddof, keepdims=keepdims, corrected=corrected
        )

    def std(self, axis=None, ddof=0, keepdims=False, corrected=False):
        """See `aesara.tensor.math.std`."""
        return at.math.std(
            self, axis=axis, ddof=ddof, keepdims=keepdims, corrected=corrected
        )

    def min(self, axis=None, keepdims=False):
        """See `aesara.tensor.math.min`."""
        return at.math.min(self, axis, keepdims=keepdims)

    def max(self, axis=None, keepdims=False):
        """See `aesara.tensor.math.max`."""
        return at.math.max(self, axis, keepdims=keepdims)

    def argmin(self, axis=None, keepdims=False):
        """See `aesara.tensor.math.argmin`."""
        return at.math.argmin(self, axis, keepdims=keepdims)

    def argmax(self, axis=None, keepdims=False):
        """See `aesara.tensor.math.argmax`."""
        return at.math.argmax(self, axis, keepdims=keepdims)

    def nonzero(self, return_matrix=False):
        """See `aesara.tensor.basic.nonzero`."""
        return at.nonzero(self, return_matrix=return_matrix)

    def nonzero_values(self):
        """See `aesara.tensor.basic.nonzero_values`."""
        return at.nonzero_values(self)

    def sort(self, axis=-1, kind="quicksort", order=None):
        """See `aesara.tensor.sort.sort`."""
        return at.sort(self, axis, kind, order)

    def argsort(self, axis=-1, kind="quicksort", order=None):
        """See `aesara.tensor.sort.argsort`."""
        from aesara.tensor.sort import argsort

        return argsort(self, axis, kind, order)

    def clip(self, a_min, a_max):
        "See `aesara.tensor.math.clip`."
        return at.math.clip(self, a_min, a_max)

    def conj(self):
        """See `aesara.tensor.math.conj`."""
        return at.math.conj(self)

    conjugate = conj

    def repeat(self, repeats, axis=None):
        """See `aesara.tensor.basic.repeat`."""
        return at.extra_ops.repeat(self, repeats, axis)

    def round(self, mode=None):
        """See `aesara.tensor.math.round`."""
        return at.math.round(self, mode)

    def trace(self):
        return at.linalg.trace(self)

    # This value is set so that Aesara arrays will trump NumPy operators.
    __array_priority__ = 1000

    def get_scalar_constant_value(self):
        return at.basic.get_scalar_constant_value(self)

    def zeros_like(model, dtype=None):
        return at.basic.zeros_like(model, dtype=dtype)

    def ones_like(model, dtype=None):
        return at.basic.ones_like(model, dtype=dtype)

    def cumsum(self, axis=None):
        return at.extra_ops.cumsum(self, axis)

    def cumprod(self, axis=None):
        return at.extra_ops.cumprod(self, axis)

    def searchsorted(self, v, side="left", sorter=None):
        return at.extra_ops.searchsorted(self, v, side, sorter)

    def ptp(self, axis=None):
        """See `aesara.tensor.math.ptp`."""

        return at.math.ptp(self, axis)

    def swapaxes(self, axis1, axis2):
        """See `aesara.tensor.basic.swapaxes`.

        If a matrix is provided with the right axes, its transpose
        will be returned.

        """
        return at.basic.swapaxes(self, axis1, axis2)

    def fill(self, value):
        """Fill inputted tensor with the assigned value."""
        return at.basic.fill(self, value)

    def choose(self, choices, out=None, mode="raise"):
        """
        Construct an array from an index array and a set of arrays to choose
        from.

        """
        return at.basic.choose(self, choices, out=None, mode="raise")

    def squeeze(self):
        """
        Remove broadcastable dimensions from the shape of an array.

        It returns the input array, but with the broadcastable dimensions
        removed. This is always `x` itself or a view into `x`.

        """
        return at.extra_ops.squeeze(self)

    def compress(self, a, axis=None):
        """Return selected slices only."""
        return at.extra_ops.compress(self, a, axis=axis)


class TensorVariable(_tensor_py_operators, Variable):
    """
    Subclass to add the tensor operators to the basic `Variable` class.

    """

    def __init__(self, type, owner=None, index=None, name=None):
        super().__init__(type, owner=owner, index=index, name=name)
        if config.warn_float64 != "ignore" and type.dtype == "float64":
            msg = (
                "You are creating a TensorVariable "
                "with float64 dtype. You requested an action via "
                "the Aesara flag warn_float64={ignore,warn,raise,pdb}."
            )
            if config.warn_float64 == "warn":
                # Get the user stack. We don't want function inside the
                # tensor and graph directory to be shown to the user.
                x = tb.extract_stack()
                nb_rm = 0
                while x:
                    file_path = x[-1][0]
                    rm = False
                    for p in [
                        "aesara/tensor/",
                        "aesara\\tensor\\",
                        "aesara/graph/",
                        "aesara\\tensor\\",
                    ]:
                        if p in file_path:
                            x = x[:-1]
                            nb_rm += 1
                            rm = True
                            break
                    if not rm:
                        break
                warnings.warn(msg, stacklevel=1 + nb_rm)
            elif config.warn_float64 == "raise":
                raise Exception(msg)
            elif config.warn_float64 == "pdb":
                import pdb

                pdb.set_trace()


TensorType.Variable = TensorVariable


class TensorConstantSignature(tuple):
    """
    A Signature object for comparing TensorConstant instances.

    An instance is a pair: (Type instance, ndarray).

    """

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        try:
            (t0, d0), (t1, d1) = self, other
        except Exception:
            return False

        # N.B. compare shape to ensure no broadcasting in ==
        if t0 != t1 or d0.shape != d1.shape:
            return False

        self.no_nan  # Ensure has_nan is computed.
        # Note that in the comparisons below, the elementwise comparisons
        # come last because they are the most expensive checks.
        if self.has_nan:
            other.no_nan  # Ensure has_nan is computed.
            return (
                other.has_nan
                and self.sum == other.sum
                and (self.no_nan.mask == other.no_nan.mask).all()
                and
                # Note that the second test below (==) may crash e.g. for
                # a single scalar NaN value, so we do not run it when all
                # values are missing.
                (self.no_nan.mask.all() or (self.no_nan == other.no_nan).all())
            )
        else:
            # Simple case where we do not need to worry about NaN values.
            # (note that if there are NaN values in d1, this will return
            # False, which is why we do not bother with testing `other.has_nan`
            # here).
            return (self.sum == other.sum) and np.all(d0 == d1)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        t, d = self
        return hash((type(self), t, d.shape, self.sum))

    def aesara_hash(self):
        _, d = self
        return hash_from_ndarray(d)

    def _get_sum(self):
        """Compute sum of non NaN / Inf values in the array."""
        try:
            return self._sum
        except AttributeError:
            self._sum = self.no_nan.sum()
            # The following 2 lines are needede as in Python 3.3 with NumPy
            # 1.7.1, numpy.ndarray and numpy.memmap aren't hashable.
            if isinstance(self._sum, np.memmap):
                self._sum = np.asarray(self._sum).item()
            if self.has_nan and self.no_nan.mask.all():
                # In this case the sum is not properly computed by numpy.
                self._sum = 0
            if np.isinf(self._sum) or np.isnan(self._sum):
                # NaN may happen when there are both -inf and +inf values.
                if self.has_nan:
                    # Filter both NaN and Inf values.
                    mask = self.no_nan.mask + np.isinf(self[1])
                else:
                    # Filter only Inf values.
                    mask = np.isinf(self[1])
                if mask.all():
                    self._sum = 0
                else:
                    self._sum = np.ma.masked_array(self[1], mask).sum()
                # At this point there should be no more NaN.
                assert not np.isnan(self._sum)
        return self._sum

    sum = property(_get_sum)

    def _get_no_nan(self):
        try:
            return self._no_nan
        except AttributeError:
            nan_mask = np.isnan(self[1])
            if nan_mask.any():
                self._no_nan = np.ma.masked_array(self[1], nan_mask)
                self.has_nan = True
            else:
                self._no_nan = self[1]
                self.has_nan = False
        return self._no_nan

    no_nan = property(_get_no_nan)


def get_unique_value(x: TensorVariable) -> Optional[Number]:
    """Return the unique value of a tensor, if there is one"""
    if isinstance(x, Constant):
        data = x.data

        if isinstance(data, np.ndarray) and data.ndim > 0:
            flat_data = data.ravel()
            if flat_data.shape[0]:
                if (flat_data == flat_data[0]).all():
                    return flat_data[0]

    return None


class TensorConstant(TensorVariable, Constant):
    """Subclass to add the tensor operators to the basic `Constant` class.

    To create a TensorConstant, use the `constant` function in this module.

    """

    def __init__(self, type, data, name=None):
        Constant.__init__(self, type, data, name)

    def __str__(self):
        unique_val = get_unique_value(self)
        if unique_val is not None:
            val = f"{self.data.shape} of {unique_val}"
        else:
            val = f"{self.data}"
        if len(val) > 20:
            val = val[:10] + ".." + val[-10:]

        if self.name is not None:
            name = self.name
        else:
            name = "TensorConstant"
        return "%s{%s}" % (name, val)

    def signature(self):
        return TensorConstantSignature((self.type, self.data))

    def equals(self, other):
        # Override Constant.equals to allow to compare with
        # numpy.ndarray, and python type.
        if isinstance(other, (np.ndarray, int, float)):
            # Make a TensorConstant to be able to compare
            other = at.basic.constant(other)
        return (
            isinstance(other, TensorConstant) and self.signature() == other.signature()
        )

    def __copy__(self):
        # We need to do this to remove the cached attribute
        return type(self)(self.type, self.data, self.name)

    def __deepcopy__(self, memo):
        # We need to do this to remove the cached attribute
        return type(self)(
            copy.deepcopy(self.type, memo),
            copy.deepcopy(self.data, memo),
            copy.deepcopy(self.name, memo),
        )


TensorType.Constant = TensorConstant
